#
# Black and White OBJ creature model exporter V0.5
# ------------------------------------------------
#
#
# This script takes a Black and White 2 creature model and exports an OBJ with
# vertex groups. All four of the creature model files must be in the same folder!
# .CCS .CSK .CBN .CHA
# Many parts of the files use constant offsets or header sizes. Hard-coded
# addresses and offsets in this code reflect that.
#
# USAGE: Argument should be any *ONE* of the four creature files.
#
# You must have Python installed for this to work. Written around Python 2.6.
# I am not a programmer.. data structures and solutions will generally
# not be efficiently written, as well as not using functions
#
# I found that many 3D programs do not handle multi-group OBJ entries correctly.
# In fact, it would seem that MOST do not. OBJ lines such as "g thigh leg knee"
# should mean that the following faces are part of those three groups (seperated
# by spaces). This will be mistakenly read as a single group called
# 'thigh leg knee'. There is nothing to be done about this. As long as these names
# are kept correctly, they should be imported correctly, but it will make group
# naming of added verticies annoying. The real solution is to use a format other
# than OBJ.
#    ***NOTE*** As it turns out, it isn't practical to do any mesh/group
# manipulation with the OBJ format. Groups are written out, but the OBJ groups
# are ignored by the OBJ import script.
#    Also, programs may not alphabatize the groups. Since groups will not appear
# in order in the OBJ, they are displayed in an annoying out-of-order way in the
# programs. To solve this, each group is declared in order first, directly before
# the faces are written.
#
# The CCS model seems to be the original 'manifold' mesh (for the most part). The
# CSK model is the model actually used by the game engine with duplicate verticies
# on the UV seams. Since the CSK model has duplcate vertices which do not share
# any faces, this will create discontinuities along the UV seams, which may be
# displayed oddly in some 3D programs. However, the creature morphs only exist as
# displacements of the CSK model, and so can only be extracted when selecting to
# export the CSK model.
#    Both models WILL have the same number of faces
#    A method to tie the morphs to the CCS model may be created in the future.
#
#
#
#	V.05
#	 - Significant changes. Most of the script was entirely re-written to accomodate
#	   new knowledge of the CCS file.
#	 - Added in CCS model functionality.
#	 - Added in grouping through CCS group definitions
#	 - Generally removed CSK welding ability since CCS model extraction does the same
#	   thing and is a superior result. The import script may need an 'un-welder' though
#	V.04
#     - Significant cleanup. Switched to reading whole file into memory instead of reading
#       tiny chunks off the disk continually.
#     - added more commentary and split into sections
#     - Fixed welding UV problem. Needed to specify > 1 UV coordinate per welded vertex
#       TO DO: Also related to welding, need to average normal at welded vertices, of each
#              vertex at that location, not just copy a value from one of the verticies.
#	V0.3
#	 - Fixed a paranthesis problem when negating normals while extracting morphs that caused
#	   the normals to not be negated and to be calculated incorrectly
#	V0.2
#     - Fixed several errata in comments and code.
#	 - Added the welding functionality
#
# int, (arrayOfInts), [int, int, int]
# CSKMats = [ array-count, [(partarray), [num-parts,index,count , ..repeat..]
#                           (partarray)   num-parts,index,count
#                           (partarray)]  num-parts,index,count]
#
# 
# float, float, float
# CSKVerts / CCSVerts / CSKNormals / CCSNormals =
#   [ (X, Y, Z) , (X, Y, Z), ... ]
#
# 
# All are floats
# CSKTANBI = 
#   [ ( Tan1X, Tan1Y, Tan1Z, BiTan1X, BiTan1Y, BiTan1Z,
#       Tan2X, Tan2Y, Tan2Z, BiTan2X, BiTan2Y, BiTan2Z ), ... ]
#
#
# float, float
# CSKUV = 
#   [ (U, V), (U, V), ... ]
#
#
# word, word, word, word (padding), float, float, float
# CSKBoneWeights = 
#   [ ( boneNum1, boneNum2, boneNum3, 0,
#       boneweight1, boneweight2, boneweight3), ... ]
#
#
# int, int, int
# CSKFaces / CCSFaces = 
#   [ (V1, V2, V3), (V1, V2, V3), ... ]
#
#
# int, int, int, int
# CSKEdgeConn = 
#   [ ( EdgeFaceIndex, CCSVertIndex, OppositeEdge, NextEdge), ... ]
#
#
# all ints
# CSKConn256 = 
#   [ ( Edge1, Edge2, Edge3, material, part-array, CSKVert1, CSKVert2, CSKVert3), ... ]
#
#
# All ints
# CCSIKChains = 
#   [ [ part1, part2, part3, ...], [ part1, part2, part3, ...], ... ]
#
#
# all floats
# CCSBoundingBox = 
#   [ (X1, Y1, Z1, X2, Y2, Z2), ... ]
#
#
# all ints
# CCSFaceGroupArrays = 
#   [ [partCount, part1, part2, ...], [partCount, part1, part2, ... ], ... ]
#
#
# TO DO: It's possible that the rotation order of the edge faces is the same as
#        the CSK model faces.. if so, only need 3 tests instead of 6.

import struct
import sys
import os




################################################
#           Variables Used                     #
################################################

#Values considered as 'constants' for the model
#either read in or calculated.
CSK_NUM_MATS = 0
CSK_BODY_PART_MAX = 0
CSK_FACE_ARRAY_ENTRY_COUNT = 0
CSK_VERT_META_ARRAY_SUBARRAY_SIZE = 0
CSK_VERT_META_ARRAY_SIZE = 0
CSK_BASE_VERTEX_ARRAY_SUB_ARRAY_SIZE = 0
CSK_BASE_VERTEX_ARRAY_SIZE = 0
CSK_MODEL_VERTEX_COUNT = 0
CSK_MORPH_COUNT = 0
CSK_FACE_COUNT = 0
CSK_EDGE_COUNT = 0


CCS_IK_ARRAY_NUM = 0
CCS_IK_ARRAY_ENTRY_NUM = 0
CCS_VERTEX_COUNT = 0
CCS_FACE_COUNT = 0
CCS_BODY_PART_COUNT = 0


#"constants" changed within various sections
MORPH_VERTEX_ARRAY_SUB_ARRAY_SIZE = 0
MORPH_VERTEX_ARRAY_SIZE = 0
NUM_PART_ARRAYS = 0



#CSK data structure addresses
#----------------------------
#Face array address
CSKfaceArrayLoc = 0
#UV/Meta array address
CSKUVArrayLoc = 0
#Vertex array index
CSKVertexArrayLoc = 0
#Where the morph sections begin
CSKMorphSectionBegin = 0




#Various indecies
#----------------
i = 0
j = 0
k = 0
m = 0
h = 0

#scratch space
scratchvalue = ''

#holds temporary values
valuebuffer = ''

#Holds the various strings as written to the OBJ files
#For extrated morphs, this must be kept after creating
#the CSK base model since they will need to be written
#to the morph files
vertexString = ''
normalString = ''
UVString = ''
faceString = ''

#Temporary array used during OBJ group creation
faceUsed = []


#Hold entire contents of CSK in memory
CSKFilebuffer = ''
#Holds a current address into the CSKFilebuffer used for reading
#variable length portion of the file
CSKFilebufferPT = 0



#Hold entire contents of CCS in memory
CCSFilebuffer = ''
#Holds a current address into the CCSFilebuffer used for reading
#variable length portion of the file
CCSFilebufferPT = 0




#Filename vars
#-------------
#filename of the model file given
filename = ''
#filename of the model file given without extension
filenamenoext = ''
#extension only of the model filename
filenameext = ''
#full path to the model file, plus the filename, but no extension
pathtomodelnoext = ''
#The directory where the files are with a '/' at the end
filedir = ''
#File handle of file currently being read
orfile = ''
#File handle of file currently being written
owfile = ''



#Option related Vars
#-------------------
userInput =''
#Negate Vertex Z
NVZ = 1
#Negate UV's (misnomer.. not really negating, just re-arranging)
NUV = 0
#Negate Normal Z
NNZ = 1
#Which model file to export 0 = CSK, 1 = CCS
modelExpChoice = 0
#Should the morphs be extracted? (only with CSK)
extractmorphs = 0
#Should the faces be exported? Allows for only exporting vertices
extractFaces = 0


#CSK
#------------

#Holds array describing the 'material' sections and associated
#body-part arrays.
CSKMats = []

#Holds the vertex tuples for model in the CSK file.
#There should be one index per vertex in the model
CSKVerts = []

#Holds the normals for each vertex in the CSK model.
#The index corresponds to a vertex
CSKNormals = []

#Holds the tangents and bitangents found in the CSK file.
#There are four entries:
#1st Tangent, 1st Bitangent, 2nd tangent, 2nd bitangent
CSKTANBI = []

#Holds the UV data for the CSK model
#There should be one index per vertex in the model
CSKUV = []

#Holds bone association/weight for each vertex. This info
#is found in the same array as the UV info.
CSKBoneWeights = []

#Holds the Faces in the CSK model. Reconstructed from the
#triangle strip array
CSKFaces = []

#Holds the unpacked data from the section called
#EdgeConnectivity
CSKEdgeConn = []

#Holds the unpacked data from the section called
#Connectivity256Const
CSKConn256 = []

#The faces defined in the Edge sections do not match
#the order of the faces as defined in the actual CSK
#model. This holds an association of the two face sets
#INDEX = CSK model face index
#VALUE = Edge face index
CSKModelFaceToEdgeFace = []

#Using the edge sections in the CSK, an association of the CSK
#verticies to CCS vertices. Note that this is NOT a directly
#reversible process. A single CCS vertex will equate to
#multiple CSK verticies.
#INDEX = CSK model vertex index
#VALUE = CCS model vertex index
CSKtoCCSVerts = []

#Using the edge sections in the CSK, as association of the
#CSK faces to CCS faces.
#INDEX = CSK model face index
#VALUE = CCS model face index
CSKtoCCSFaces = []




#CCS
#------------

#Holds the info from what is currently guessed to be 'IK chains'
CCSIKChains = []

#Holds the X,Y,Z tuples for the vertices from the CCS model
CCSVerts = []

#Holds the body part association for each vertex in the CCS model.
#The index corresponds to a vertex and the value specifies the
#body-part number the vertex belongs to
#INDEX = CCS model vertex index
#VALUE = body-part number
CCSVertPartArray = []

#Holds the X,Y,Z values of normals for each vertex in the CCS model.
#The index corresponds to a vertex
CCSNormals = []

#Holds the face info for the CCS. Each index is a face, the value is
#a tuple holding indexes of the vertices that make up the face
CCSFaces = []

#Holds the bounding boxes for the CCS model. The index corresponds to
#a body-part number. The values are two X,Y,Z coordinates that define
#two points. These two points define a 3D box that should encompass
#the body-part
CCSBoundingBox = []

#Array of arrays that holds the face-to-bodypart association for the
#CCS mode. The index corresponds to a body part. The sub-array is a
#list of face index numbers that indicate that the face belongs to that
#body part. The FIRST value in the sub-array is a count of how many
#faces are in the sub-array. A face can belong to more than on array.
CCSFaceGroupArrays = []

#Holds a calculated "reverse-lookup" for the part arrays contained
#in the CCS. Each index corresponds to a face and the values indicate
#which body parts the face belongs to. Each face can belong to up to 3
#different parts
#INDEX = CCS model face index
#VALUES = body part numbers (up to three)
CCSFaceToPart = []

#Holds an association of the CCS faces to the CSK model faces. This is
#a 'reverse-lookup' of the CSKtoCCSFaces array above.
#INDEX = CCS model face index
#VALUE = CSK model face index
CCStoCSKFaces = []


###########################################
#        Arguments                        #
###########################################
if len(sys.argv) < 2:
	print "Missing Argument:  Any creature model file"
	print "  ( CCS CBN CSK CHA )"
	print
	raise SystemExit

filename = os.path.basename(sys.argv[1])
filedir = os.path.dirname(sys.argv[1])
if filedir != "":
	filedir += "/"
filenamenoext, filenameext = os.path.splitext(filename)
pathtomodelnoext = filedir+filenamenoext
if not os.path.isfile(filedir+filenamenoext+filenameext):
	print
	print "ERROR:File in argument does not exist"
	print
	raise SystemExit
if not os.path.isfile(pathtomodelnoext+".ccs"):
	print
	print "ERROR: CCS file does not exist"
	print
	raise SystemExit
if not os.path.isfile(pathtomodelnoext+".cha"):
	print
	print "ERROR: CHA file does not exist"
	print
	raise SystemExit
if not os.path.isfile(pathtomodelnoext+".csk"):
	print
	print "ERROR: CSK file does not exist"
	print
	raise SystemExit
if not os.path.isfile(pathtomodelnoext+".cbn"):
	print
	print "ERROR: CBN file does not exist"
	print
	raise SystemExit
################################################
################################################








################################################
#           Get Extraction Options             #
################################################
print
print
print "CTRL-C at any point to break this script."
print
print
print
print "Extract which model?"
print "--------------------"
print "  1) CSK - duplicate verts"
print "  2) CCS - no duplicate verts"
print
print "IMPORTANT: Currently, only a CSK model"
print "export may be re-imported for the game."
print "If you intend to re-import for the game,"
print "you MUST choose the CSK model here."
print 
print "Note that morphs may only currently be"
print "extracted when selecting the CSK model."
print
while userInput != "1" and userInput != "2":
	userInput = raw_input('--> ')
	if userInput != "1" and userInput != "2":
		print "Input is not correct"
if userInput == "2":
	modelExpChoice = 1

userInput = ''
print
print
print
print "Should Vertex Z be inverted?"
print "  1) Y"
print "  2) N"
print
while userInput != "1" and userInput != "2":
	userInput = raw_input('--> ')
	if userInput != "1" and userInput != "2":
		print "Input is not correct"
if userInput == "1":
	NVZ = -1

userInput = ''
print
print
print "Should Normal vector Z be inverted?"
print "  1) Y"
print "  2) N"
print
while userInput != "1" and userInput != "2":
	userInput = raw_input('--> ')
	if userInput != "1" and userInput != "2":
		print "Input is not correct"
if userInput == "1":
	NNZ = -1

userInput = ''
print
print
print "Should UV's be arranged to match .dds texture files?"
print "If this is not done, the texture files will not inherently"
print "match exported model. See readme.txt for explanation."
print "  1) Y"
print "  2) N"
print
while userInput != "1" and userInput != "2":
	userInput = raw_input('--> ')
	if userInput != "1" and userInput != "2":
		print "Input is not correct"
if userInput == "1":
	NUV = 1

userInput = ''
print
print
print "Extract faces and verticies"
print "or only extract vertices?"
print "  1) Faces and Verts"
print "  2) Verts only"
print
while userInput != "1" and userInput != "2":
	userInput = raw_input('--> ')
	if userInput != "1" and userInput != "2":
		print "Input is not correct"
if userInput == "1":
	extractFaces = 1

###Curently, only extract morphs if the CSK model is
###exported. May add in special welding function later
###Problem is dealing with eyes where CCS model keeps
###duplicate verts still, can't do original welding function
if modelExpChoice == 0:
	userInput = ''
	print
	print
	print "Should the morphs be extracted?"
	print "The output filename will be"
	print "modelname-morphname.obj."
	print "These will be extracted as fully"
	print "independant models, not"
	print "displacement values."
	print "  1) Yes"
	print "  2) No"
	print
	while userInput != "1" and userInput != "2":
		userInput = raw_input('--> ')
		if userInput != "1" and userInput != "2":
			print "Input is not correct"
	if userInput == "1":
		extractmorphs = 1


print
print "----------------------------"
print
################################################
################################################







###########################################
#        open CSK and create base OBJ     #
###########################################
try:
	orfile = open(pathtomodelnoext+".csk", 'rb')
except:
	print
	print "ERROR:Cannot open CSK file for reading"
	print "Why? File seems to exist. Persmission problems?"
	print
	raise SystemExit


try:
	owfile = open(pathtomodelnoext+".obj", 'w')
except:
	print
	print "ERROR:Cannot open file for writing."
	print "Disk Full? Directory read-only?"
	print
	raise SystemExit
################################################
################################################




################################################
#           Read CSK file and get header       #
################################################
print "Reading CSK . . .",

orfile.seek(0)
CSKFilebuffer = orfile.read()

###Skip the file header info
CSKFilebufferPT = 52

###Get the maximum body part count
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
CSKFilebufferPT += 4
CSK_BODY_PART_MAX = struct.unpack_from('<i', valueBuffer)[0]


###If the first section is not 'radeon', Skip it. In that case,
###Gets the size of the first section, then jumps ahead that
###number of bytes.
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+32]
CSKFilebufferPT += 32
if valueBuffer[0] != "r" and valueBuffer[0] != "R":
	valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
	CSKFilebufferPT += 4
	scratchvalue = struct.unpack_from('<i', valueBuffer)[0]
	###Skip the section
	CSKFilebufferPT += scratchvalue
	###Skip the second section's header
	CSKFilebufferPT += 36
else:
	###If the section IS 'radeon', skip the section-size value
	CSKFilebufferPT += 4
################################################
################################################





################################################
#           Read CSK material sections         #
################################################
arraysRead = 0


###The Mat array data structure is documented at the top of this file.
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
CSKFilebufferPT += 4
CSK_NUM_MATS = struct.unpack_from('<i', valueBuffer)[0]

i = 0
while i != CSK_NUM_MATS:
	CSKFilebufferPT += 516
	valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
	CSKFilebufferPT += 4
	NUM_PART_ARRAYS = struct.unpack_from('<i', valueBuffer)[0]
	CSKMats.append(int(NUM_PART_ARRAYS))
	CSKMats.append([])
	CSKMats.append([])
	j = 0
	while j != NUM_PART_ARRAYS:
		valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
		CSKFilebufferPT += 4
		PART_ARRAY_COUNT = struct.unpack_from('<i', valueBuffer)[0]
		CSKMats[(i*3)+2].append(PART_ARRAY_COUNT)
		### Read the array and the two count numbers
		valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+(PART_ARRAY_COUNT * 4)]
		CSKFilebufferPT += (PART_ARRAY_COUNT * 4)
		CSKMats[(i*3)+1].append(struct.unpack_from('<'+str(PART_ARRAY_COUNT)+'i', valueBuffer))
		valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+8]
		CSKFilebufferPT += 8
		CSKMats[(i*3)+2].append(struct.unpack_from('<i', valueBuffer)[0])
		CSKMats[(i*3)+2].append(struct.unpack_from('<2i', valueBuffer)[1])
		arraysRead += 1
		j += 1
	i += 1
################################################
################################################




################################################
#           Skip CSK Face Array                #
################################################
###Now at the beginning of the Face array.
###For convenience, this should be done later.
###Keep the address
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
CSKFilebufferPT += 4
CSK_FACE_ARRAY_ENTRY_COUNT = struct.unpack_from('<i', valueBuffer)[0]
CSKfaceArrayLoc = CSKFilebufferPT
###The entries in this array are words (2 bytes)
CSKFilebufferPT += (CSK_FACE_ARRAY_ENTRY_COUNT * 2)
################################################
################################################




################################################
#           Skip CSK UV/Meta array             #
################################################
###Now at the beginning of the UV arrays. These
###hold UV data and bone associations/weight for
###each vertex. Also skip it for now.
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
CSKFilebufferPT += 4
CSK_VERT_META_ARRAY_SUBARRAY_SIZE = struct.unpack_from('<i', valueBuffer)[0]
if CSK_VERT_META_ARRAY_SUBARRAY_SIZE != 28:
	print
	print "ERROR: UV array in unknown format"
	print "Each sub-array should be 28-bytes"
	print
	orfile.close()
	owfile.close()
	raise SystemExit
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
CSKFilebufferPT += 4
CSK_VERT_META_ARRAY_SIZE = struct.unpack_from('<i', valueBuffer)[0]
CSKUVArrayLoc = CSKFilebufferPT
CSKFilebufferPT += CSK_VERT_META_ARRAY_SIZE
################################################
################################################





################################################
#           Read CSK vertex array              #
################################################
###The vertex array contains (X,Y,Z), Normal,
###tangent, and bitangent info per vertex. There
###Are two tan/bitan pairs per vertex (12 numbers)
###but I don't know what effect the second pair has


###Begin extracting the vertices
CSKFilebufferPT += 32
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
CSKFilebufferPT += 4
CSK_BASE_VERTEX_ARRAY_SUB_ARRAY_SIZE = struct.unpack_from('<i', valueBuffer)[0]
if CSK_BASE_VERTEX_ARRAY_SUB_ARRAY_SIZE != 72:
	print
	print "ERROR:Vertex array in unknown format"
	print "Each sub-array should be 72-bytes"
	print
	orfile.close()
	owfile.close()
	raise SystemExit
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
CSKFilebufferPT += 4
CSK_BASE_VERTEX_ARRAY_SIZE = struct.unpack_from('<i', valueBuffer)[0]
CSK_MODEL_VERTEX_COUNT = CSK_BASE_VERTEX_ARRAY_SIZE/CSK_BASE_VERTEX_ARRAY_SUB_ARRAY_SIZE

###Remember the position of the vertex array
CSKVertexArrayLoc = CSKFilebufferPT

valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+CSK_BASE_VERTEX_ARRAY_SIZE]
CSKFilebufferPT += CSK_BASE_VERTEX_ARRAY_SIZE
j = CSK_BASE_VERTEX_ARRAY_SIZE
i = 0
while i < j:
	CSKVerts.append(struct.unpack_from('<3f', valueBuffer[i:i+12:1]))
	CSKNormals.append(struct.unpack_from('<3f', valueBuffer[i+12:i+24:1]))
	CSKTANBI.append(struct.unpack_from('<12f', valueBuffer[i+24:i+72:1]))
	i += 72



################################################
#           Read CSK UV/Meta array             #
################################################
#The UV array contains UV data and bone
#association/weights per vertex.

###First, save the morph section location. If the CSK model is
###being exported and the morphs are going to be extracted,
###it will be necessary to go back to this address
CSKMorphSectionBegin = CSKFilebufferPT


###Go back and get the UV texture coordinate array
CSKFilebufferPT = CSKUVArrayLoc
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+CSK_VERT_META_ARRAY_SIZE]
CSKFilebufferPT += CSK_VERT_META_ARRAY_SIZE

j = CSK_VERT_META_ARRAY_SIZE
i = 0
while i != j:
	CSKUV.append(struct.unpack_from('<2f', valueBuffer[i:i+8:1]))
	CSKBoneWeights.append( ( struct.unpack('<h', valueBuffer[i+8:i+10])[0], struct.unpack('<f', valueBuffer[i+16:i+20])[0], struct.unpack('<h', valueBuffer[i+10:i+12])[0], struct.unpack('<f', valueBuffer[i+20:i+24])[0], struct.unpack('<h', valueBuffer[i+12:i+14])[0], struct.unpack('<f', valueBuffer[i+24:i+28])[0] ) )
	i += 28





################################################
# Reconstruct faces from CSK face index array  #
################################################
#Holds the entire face array as an unpacked, parsable data structure
faceTuple = ''
numArrays = 0
endArrayEntryNum = 0


CSKFilebufferPT = CSKfaceArrayLoc

###Each entry is a word, 2-bytes
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+(CSK_FACE_ARRAY_ENTRY_COUNT * 2)]
CSKFilebufferPT += CSK_FACE_ARRAY_ENTRY_COUNT * 2

#create an array of *words* of the entire face array
faceTuple = struct.unpack_from('<' + str(CSK_FACE_ARRAY_ENTRY_COUNT) + 'h', valueBuffer)


i = 0
while i != CSK_NUM_MATS:
	###numArrays will be the number of part arrays in this Material
	numArrays = CSKMats[i*3]
	j = 0
	while j != numArrays:
		###k will be the beginning index of this part-array's listings in the face array
		k = CSKMats[(i*3)+2][(j*3)+1]
		###endArrayEntryNum will be the last index number in the face array for this
		###part-array
		endArrayEntryNum = CSKMats[(i*3)+2][(j*3)+2] + k
		###Faces will be triplets, (index-1, index, index+1). For index-1 and index+1 to
		###exist correctly, the reading must start at k+1 and end at endArrayEntryNum-1
		k = k + 1
		endArrayEntryNum = endArrayEntryNum - 1
		h = 1
		while k != endArrayEntryNum:
			###As long as any of the numbers in the triplet are not the same...
			if faceTuple[k] != faceTuple[k-1] and faceTuple[k] != faceTuple[k+1] and faceTuple[k-1] != faceTuple[k+1]:
				if h:
					CSKFaces.append( ( faceTuple[k-1] , faceTuple[k] , faceTuple[k+1] ) )
				else:
					CSKFaces.append( ( faceTuple[k] , faceTuple[k-1] , faceTuple[k+1] ) )
			h = not h
			k += 1
		j += 1
	i += 1

CSK_FACE_COUNT = len(CSKFaces)
################################################
################################################





################################################
#          Read CSK edge info                  #
################################################
edgeArraySize = 0

###The morph section of the CSK will have to be skipped first.
###It will take some effort to simply find the beinning of the
###edge arrays..
###First, go back to the beinning of the morph secion...
CSKFilebufferPT = CSKMorphSectionBegin
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
CSKFilebufferPT += 4
CSK_MORPH_COUNT = struct.unpack_from('<i', valueBuffer)[0]

i = 0
while i < CSK_MORPH_COUNT:
	###morph name
	CSKFilebufferPT += 32
	###Skip two counts, get the sub-array size
	valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+8]
	CSKFilebufferPT += 8
	MORPH_VERTEX_ARRAY_SUB_ARRAY_SIZE = struct.unpack_from('<i', valueBuffer)[0]

	###Skip bitwise array. 1 byte per vertex
	CSKFilebufferPT += CSK_MODEL_VERTEX_COUNT

	###Get morph array size
	valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
	CSKFilebufferPT += 4
	MORPH_VERTEX_ARRAY_SIZE = struct.unpack_from('<i', valueBuffer)[0]

	###Skip morph array
	CSKFilebufferPT += MORPH_VERTEX_ARRAY_SIZE

	i += 1

###Should now be at the start of the edge sections
###Skip header and size dword
CSKFilebufferPT += 36

###Get edge count 
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
CSKFilebufferPT += 4
CSK_EDGE_COUNT = struct.unpack_from('<i', valueBuffer)[0]

###Each edge subarray is 16 bytes (4 dwords)
edgeArraySize = CSK_EDGE_COUNT*16

valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+edgeArraySize]
CSKFilebufferPT += edgeArraySize


i = 0
while i < edgeArraySize:
	CSKEdgeConn.append( struct.unpack_from('<4i', valueBuffer[i:i+16]) )
	i += 16

###Now, skip the '96' section
CSKFilebufferPT += 32
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
CSKFilebufferPT += 4
edgeArraySize = struct.unpack('<i', valueBuffer)[0]
CSKFilebufferPT += edgeArraySize

###Get the '256' section
CSKFilebufferPT += 32
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
CSKFilebufferPT += 4
edgeArraySize = struct.unpack('<i', valueBuffer)[0]
###Skip the face count.. already know it
CSKFilebufferPT += 4
###The edgecount array size included the face count number.. don't want
###this in the count, so sub 4.
edgeArraySize = edgeArraySize - 4

###Each sub-array is 32 bytes long (8 ints)
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+edgeArraySize]
CSKFilebufferPT += edgeArraySize

i = 0
while i < edgeArraySize:
	CSKConn256.append( struct.unpack_from('<8i', valueBuffer[i:i+32]) )
	i += 32

valueBuffer = ''

################################################
################################################



orfile.close()

print " done"

################################################
################################################
################################################
################################################






################################################
#               Read CCS file                  #
################################################
try:
	orfile = open(pathtomodelnoext+".ccs", 'rb')
except:
	print
	print "ERROR:Cannot open CCS file for reading"
	print "Why? File seems to exist. Persmission problems?"
	print
	owfile.close()
	raise SystemExit

print "Reading CCS . . .",

orfile.seek(0)
CCSFilebuffer = orfile.read()
CCSFilebufferPT = 0



###Skip CCS Header
CCSFilebufferPT += 44


################################################
#           Read CCS IK                        #
################################################

###Skip the IK entry count number
CCSFilebufferPT += 4


valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+4]
CCSFilebufferPT += 4
CCS_IK_ARRAY_NUM = struct.unpack('<i', valueBuffer)[0]



valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+4]
CCSFilebufferPT += 4
CCS_IK_ARRAY_ENTRY_NUM = struct.unpack('<i', valueBuffer)[0]


i = 0
while i < CCS_IK_ARRAY_NUM:
	j = 0
	CCSIKChains.append([])
	while j < CCS_IK_ARRAY_ENTRY_NUM:
		valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+4]
		CCSFilebufferPT += 4
		CCSIKChains[i].append(struct.unpack('<i', valueBuffer)[0])
		j += 1
	i += 1
################################################
################################################



################################################
#           Read CCS vertex info               #
################################################
valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+4]
CCSFilebufferPT += 4
CCS_VERTEX_COUNT = struct.unpack('<i', valueBuffer)[0]


i = 0
while i < CCS_VERTEX_COUNT:
	valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+12]
	CCSFilebufferPT += 12
	CCSVerts.append( ( struct.unpack('<3f', valueBuffer)[0], struct.unpack('<3f', valueBuffer)[1], struct.unpack('<3f', valueBuffer)[2] ) )
	i += 1
################################################
################################################




################################################
#  Read CCS Vertex-to-part index array         #
################################################
i = 0
while i < CCS_VERTEX_COUNT:
	valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+4]
	CCSFilebufferPT += 4
	CCSVertPartArray.append(struct.unpack('<i', valueBuffer)[0])
	i += 1
################################################
################################################


################################################
#      Read CCS Normal array                   #
################################################
i = 0
while i < CCS_VERTEX_COUNT:
	valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+12]
	CCSFilebufferPT += 12
	CCSNormals.append( ( struct.unpack('<3f', valueBuffer)[0], struct.unpack('<3f', valueBuffer)[1], struct.unpack('<3f', valueBuffer)[2] ) )
	i += 1
################################################
################################################


################################################
#      Read CCS Face array                     #
################################################
valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+4]
CCSFilebufferPT += 4
CCS_FACE_COUNT = struct.unpack('<i', valueBuffer)[0]


###Skip the eye-definition array for now
CCSFilebufferPT += (CCS_FACE_COUNT * 4)

i = 0 
while i < CCS_FACE_COUNT:
	valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+12]
	CCSFilebufferPT += 12
	CCSFaces.append( ( struct.unpack('<3i', valueBuffer)[0], struct.unpack('<3i', valueBuffer)[1], struct.unpack('<3i', valueBuffer)[2] ) )
	i += 1
################################################
################################################



################################################
#        Read CCS bounding boxes               #
################################################
valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+4]
CCSFilebufferPT += 4
CCS_BODY_PART_COUNT = struct.unpack('<i', valueBuffer)[0]


i = 0
while i < CCS_BODY_PART_COUNT:
	valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+24]
	CCSFilebufferPT += 24
	CCSBoundingBox.append( ( struct.unpack('<6f', valueBuffer)[0], struct.unpack('<6f', valueBuffer)[1], struct.unpack('<6f', valueBuffer)[2], struct.unpack('<6f', valueBuffer)[3], struct.unpack('<6f', valueBuffer)[4], struct.unpack('<6f', valueBuffer)[5] ) )
	i += 1
################################################
################################################



################################################
#    Read CCS face group association arrays    #
################################################
facenum = 0

###initialize the 'reverse-lookup' array
i = 0
while i < CCS_FACE_COUNT:
	CCSFaceToPart.append([])
	i += 1


i = 0
while i < CCS_BODY_PART_COUNT:
	CCSFaceGroupArrays.append([])
	valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+4]
	CCSFilebufferPT += 4
	j = struct.unpack('<i', valueBuffer)[0]
	CCSFaceGroupArrays[i].append(j)
	k = 0
	while k < j:
			valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+4]
			CCSFilebufferPT += 4
			facenum = struct.unpack('<i', valueBuffer)[0]
			CCSFaceGroupArrays[i].append( facenum )
			CCSFaceToPart[facenum].append(int(i))
			k += 1
	i += 1
################################################
################################################


orfile.close()

print " done"

################################################
################################################
################################################
################################################







################################################
#                 Clean up                     #
################################################






################################################
################################################







################################################
#   create CSK to CCS relation arrays          #
################################################
###I don't there's really no way to make this less
###calculation intensive.. the vertices have a direct
###association, but the faces just have to be sought
###out one by one
edgeV1 = 0
edgeV2 = 0
edgeV3 = 0
modelV1 = 0
modelV2 = 0
modelV3 = 0
ccsTestV1 = 0
ccsTestV2 = 0
ccsTestV3 = 0
currentEdgeFace = 0
faceFound = 0

print "Calculating CSK to CCS relation",

###initialize arrays
i = 0
while i < CSK_MODEL_VERTEX_COUNT:
	CSKtoCCSVerts.append( int(-1) )
	i += 1

i = 0
while i < CSK_FACE_COUNT:
	CSKtoCCSFaces.append( int(-1) )
	CCStoCSKFaces.append( int(-1) )
	CSKModelFaceToEdgeFace.append( int(-1) )
	i += 1	


###Create vertex association first. Many vertices will be set more
###than once, but I think it takes just as much processing power to
###check for 'aready set' as it does to just set it again.
i = 0
while i < CSK_FACE_COUNT:
	CSKtoCCSVerts[CSKConn256[i][5]] = CSKEdgeConn[CSKConn256[i][0]][1]
	CSKtoCCSVerts[CSKConn256[i][6]] = CSKEdgeConn[CSKConn256[i][1]][1]
	CSKtoCCSVerts[CSKConn256[i][7]] = CSKEdgeConn[CSKConn256[i][2]][1]
	i += 1

print ".",

###Now, since the faces as described in the edge sections will not
###match the indexes (the order) of the faces in the actual model
###face arrays, an association needs to be made between the edge
###faces and the model faces
i = 0
while i < CSK_FACE_COUNT:
	faceUsed.append( int(0) )
	i += 1
i = 0
while i < CSK_FACE_COUNT:
	faceFound = 0

	###Model here refers to CSk model
	modelV1 = CSKFaces[i][0]
	modelV2 = CSKFaces[i][1]
	modelV3 = CSKFaces[i][2]

	j = 0
	while (j < CSK_FACE_COUNT) and (not faceFound):
		if faceUsed[j] == 0:
			edgeV1 = CSKConn256[j][5]
			edgeV2 = CSKConn256[j][6]
			edgeV3 = CSKConn256[j][7]
			if (edgeV1 == modelV1 and edgeV2 == modelV2 and edgeV3 == modelV3) or (edgeV1 == modelV1 and edgeV2 == modelV3 and edgeV3 == modelV2) or (edgeV1 == modelV2 and edgeV2 == modelV1 and edgeV3 == modelV3) or (edgeV1 == modelV2 and edgeV2 == modelV3 and edgeV3 == modelV1) or (edgeV1 == modelV3 and edgeV2 == modelV1 and edgeV3 == modelV2) or (edgeV1 == modelV3 and edgeV2 == modelV2 and edgeV3 == modelV1):
				CSKModelFaceToEdgeFace[i] = j
				faceFound = 1
				faceUsed[j] = 1
		j += 1
	i += 1

print ".",

###Now that the CSK face association is actually known, and the CCS verts
###corresponding to the CSK verts have been determined, an association can
###be made from the CSK faces to the CCS faces
###The loop ('i') will be tracking the CSK face index. The value currentEdgeFace
###will hold the corresponding edge face index for that CSK model's face index.
###Then the vertex indexes as indicated by that EDGE face are set (should
###actually be the same vertecies as the CSK face since they use the same CSK
###model index numbers, but possibly in a different order). Then, those CSK 
###index numbers are substituted for the equivalant CCS model vertex numbers
###and then the face with those vertecies is sought out. This is then the
###equivalent face index of the CCS model to the CSK model.
i = 0
while i < CSK_FACE_COUNT:
	faceUsed[i] = 0
	i += 1
i = 0

while i < CSK_FACE_COUNT:
	faceFound = 0

	currentEdgeFace = CSKModelFaceToEdgeFace[i]
	edgeV1 = CSKConn256[currentEdgeFace][5]
	edgeV2 = CSKConn256[currentEdgeFace][6]
	edgeV3 = CSKConn256[currentEdgeFace][7]

	modelV1 = CSKtoCCSVerts[edgeV1]
	modelV2 = CSKtoCCSVerts[edgeV2]
	modelV3 = CSKtoCCSVerts[edgeV3]



	j = 0
	while (j < CCS_FACE_COUNT) and (not faceFound):
		if faceUsed[j] == 0:
			ccsTestV1 = CCSFaces[j][0]
			ccsTestV2 = CCSFaces[j][1]
			ccsTestV3 = CCSFaces[j][2]
			if (modelV1 == ccsTestV1 and modelV2 == ccsTestV2 and modelV3 == ccsTestV3) or (modelV1 == ccsTestV1 and modelV2 == ccsTestV3 and modelV3 == ccsTestV2) or (modelV1 == ccsTestV2 and modelV2 == ccsTestV1 and modelV3 == ccsTestV3) or (modelV1 == ccsTestV2 and modelV2 == ccsTestV3 and modelV3 == ccsTestV1) or (modelV1 == ccsTestV3 and modelV2 == ccsTestV1 and modelV3 == ccsTestV2) or (modelV1 == ccsTestV3 and modelV2 == ccsTestV2 and modelV3 == ccsTestV1):
				CSKtoCCSFaces[i] = j
				CCStoCSKFaces[j] = i
				faceFound = 1
				faceUsed[j] = 1
		j += 1
	i += 1

################################################
################################################


print " done"



################################################
#               Extract CSK                    #
################################################
difParts = 0
currentNumParts = 0
currentParts = [ -1, -1, -1 ]
currentFaceIndex = 0
cskV1 = 0
cskV2 = 0
cskV3 = 0

if modelExpChoice == 0:

	print "Extracting CSK model . . .",

	###initialize face used array.
	i = 0
	while i < CSK_FACE_COUNT:
		faceUsed[i] = 0
		i += 1


	###Ready verticies
	scratchvalue = ''
	
	i = 0
	while i < CSK_MODEL_VERTEX_COUNT:
		###Remember, NVZ = inverted Z modifier. will be 1 if no inversion, -1 if inversion
		vertexString += "v " + str(format(CSKVerts[i][0], ' .8f')) + " " + str(format(CSKVerts[i][1], ' .8f')) + " " + str( format((CSKVerts[i][2]*NVZ), ' .8f') ) + "\n"
		i += 1
	vertexString += "\n"

	###Ready normals
	i = 0
	while i < CSK_MODEL_VERTEX_COUNT:
		###NVZ = inverted Z modifier. will be 1 if no inversion, -1 if inversion
		normalString += "vn " + str(format(CSKNormals[i][0], ' .8f')) + " " + str(format(CSKNormals[i][1], ' .8f')) + " " + str( format((CSKNormals[i][2]*NNZ), ' .8f') ) + "\n"
		i += 1
	normalString += "\n"

	###Ready UV coordinates
	i = 0
	while i < CSK_MODEL_VERTEX_COUNT:
		if NUV:
			if CSKUV[i][0] > 0.5:
				UVString += "vt " + str( format(((1-CSKUV[i][0])*2), ' .8f') ) + " " + str( format((1 - CSKUV[i][1]), ' .8f') ) + "\n"
			else:
				UVString += "vt " + str( format(CSKUV[i][0]*2, ' .8f') ) + " " + str( format(1 - CSKUV[i][1], ' .8f') ) + "\n"
		else:
			UVString += "vt " + str(format(CSKUV[i][0], ' .8f')) + " " + str(format(CSKUV[i][1], ' .8f')) + "\n"
		i += 1
	UVString += "\n"


	###Write out the groups in order. This fixes the group listing for programs that do not
	###alphabetize the groups
	i = 0
	while i < CSK_BODY_PART_MAX:
		faceString += "g BodyPart_" + str( int(i) ) + "\n"
		i += 1


	###Prepare faces. This takes a bit more work. Vertex groups are created from the face-association
	###arrays in the CCS file.
	###This gets a bit unintuitive in the interaction between the CCS and CSK. Generally, the two
	###different models have different vertex counts, but have the 'same' verticies and do have
	###equivalent faces
	###Since the face definitions do not change between morphs, keep this string on hand to use to
	###write out to all the morph models if they are extracted.
	i = 0
	while i < CSK_BODY_PART_MAX:
		###j will be the number of faces in this part's array. +1 since the index is starting from
		###"1" to accomodate index scheme from CCSFaceGroupArrays
		j = CCSFaceGroupArrays[i][0] + 1
		k = 1
		while k < j:
			difParts = 0
			currentFaceIndex = CCSFaceGroupArrays[i][k]
			if faceUsed[currentFaceIndex] != 1:
				faceUsed[currentFaceIndex] = 1
				cskV1 = CSKFaces[CCStoCSKFaces[currentFaceIndex]][0]
				cskV2 = CSKFaces[CCStoCSKFaces[currentFaceIndex]][1]
				cskV3 = CSKFaces[CCStoCSKFaces[currentFaceIndex]][2]

				if currentNumParts != len(CCSFaceToPart[currentFaceIndex]):
					currentNumParts = len(CCSFaceToPart[currentFaceIndex])
					currentParts = CCSFaceToPart[currentFaceIndex]
					faceString += "g "
					m = 0
					while m < currentNumParts:
						faceString += "BodyPart_" + str(CCSFaceToPart[currentFaceIndex][m]) + " "
						m += 1
					faceString += "\n"

					faceString += "f " + str(cskV1+1) + "/" + str(cskV1+1) + "/" + str(cskV1+1) + " " + str(cskV2+1) + "/" + str(cskV2+1) + "/" + str(cskV2+1) + " " + str(cskV3+1) + "/" + str(cskV3+1) + "/" + str(cskV3+1) + "\n"
				else:
					m = 0
					while m < len(CCSFaceToPart[currentFaceIndex]):
						if CCSFaceToPart[currentFaceIndex][m] != currentParts[m]:
							difParts = 1
						m += 1
					if difParts:
						currentNumParts = len(CCSFaceToPart[currentFaceIndex])
						currentParts = CCSFaceToPart[currentFaceIndex]
						faceString += "g "
						m = 0
						while m < currentNumParts:
							faceString += "BodyPart_" + str(CCSFaceToPart[currentFaceIndex][m]) + " "
							m += 1
						faceString += "\n"
						faceString += "f " + str(cskV1+1) + "/" + str(cskV1+1) + "/" + str(cskV1+1) + " " + str(cskV2+1) + "/" + str(cskV2+1) + "/" + str(cskV2+1) + " " + str(cskV3+1) + "/" + str(cskV3+1) + "/" + str(cskV3+1) + "\n"
					else:
						faceString += "f " + str(cskV1+1) + "/" + str(cskV1+1) + "/" + str(cskV1+1) + " " + str(cskV2+1) + "/" + str(cskV2+1) + "/" + str(cskV2+1) + " " + str(cskV3+1) + "/" + str(cskV3+1) + "/" + str(cskV3+1) + "\n"

			k += 1
		i += 1						


	###Write out values
	owfile.write(vertexString)
	owfile.write(normalString)
	owfile.write(UVString)
	###If the option to extract faces was selected
	if extractFaces:
		owfile.write(faceString)

	print " done"


################################################
#               Extract Morphs                 #
################################################
#This basically works the same way as the base model
#extraction. The main addition is the bitwise
#description of which vertices displace per morph.
#A shifting window of 8-bits is AND-ed with the
#current bitwise-describing byte to test for
#each vertex's displacement.
morphBitArrayCount = 0
morphVertexArrayCount = 0
loopshiftcount = 8
morphname = ''
#bitArray declared below, as a type 'buffer'
VertexMorphTuple = []

###If the morphs are to be extracted
if extractmorphs:

	###Re-open the CSK file
	try:
		orfile = open(pathtomodelnoext+".csk", 'rb')
	except:
		print
		print "ERROR:Cannot open CSK file for reading (morphs)"
		print "Why? File seems to exist. Persmission problems?"
		print
		raise SystemExit

	print "Extracting CSK model morphs . . .",

	###Go to the end of the base-morph vertex array
	CSKFilebufferPT = CSKMorphSectionBegin
	valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
	CSKFilebufferPT += 4
	CSK_MORPH_COUNT = struct.unpack_from('<i', valueBuffer)[0]
	i = 0
	while i != CSK_MORPH_COUNT:
		###This is needed since loop indices will not keep track of the 'read' (past-tense) position
		###of the morph vertex array.
		morphVertexArrayCount = 0
		morphBitArrayCount = 0
		VertexMorphTuple = []
		morphname = ''
		vertexString = "\n"
		normalString = "\n"
		###Get the name
		valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+32]
		CSKFilebufferPT += 32
		###Python string functions seem to change so often, I think it is better to avoid using them...
		j = 0
		while j != 32:
			if valueBuffer[j] != '\x00':
				morphname += valueBuffer[j]
			j += 1
################################################
#           Open morph file to write to        #
################################################
		try:
			owfile = open(pathtomodelnoext+"-"+morphname+".obj", 'w')
		except:
			print
			print "ERROR:Cannot open file for writing."
			print str(pathtomodelnoext)+"-"+str(morphname)+".obj"
			print "Disk Full? Directory read-only?"
			print
			orfile.close()
			owfile.close()
			raise SystemExit
		valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+8]
		CSKFilebufferPT += 8
		MORPH_VERTEX_ARRAY_SUB_ARRAY_SIZE = struct.unpack_from('<i', valueBuffer)[0]
		if MORPH_VERTEX_ARRAY_SUB_ARRAY_SIZE != 72:
			print
			print "ERROR:Morph Vertex array in unknown format"
			print "Should be 72-bytes per sub array"
			print
			orfile.close()
			owfile.close()
			raise SystemExit


		###There will be 1 byte per vertex. this is actulally a mistake.. the array is a bit-wise reference.
		bitArray = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+CSK_MODEL_VERTEX_COUNT]
		CSKFilebufferPT += CSK_MODEL_VERTEX_COUNT

		valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
		CSKFilebufferPT += 4
		###Since the morph vertex array will not include all of the vertices, this number is important..
		MORPH_VERTEX_ARRAY_SIZE = struct.unpack_from('<i', valueBuffer)[0]

		valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+MORPH_VERTEX_ARRAY_SIZE]
		CSKFilebufferPT += MORPH_VERTEX_ARRAY_SIZE

		j = MORPH_VERTEX_ARRAY_SIZE
		k = 0
		###Read in and process the morph vertex array.. create a single, large tuple
		while k != j:
			VertexMorphTuple.append(struct.unpack_from('<18f', valueBuffer[k:k+MORPH_VERTEX_ARRAY_SUB_ARRAY_SIZE:1]))
			k += MORPH_VERTEX_ARRAY_SUB_ARRAY_SIZE
		j = 0
		###This loop will parse the bitwise references by bit-wise AND'ing a shifting mask with the byte
		###read, in a inner loop of 8-count (8 bits per byte). Note that the bit on the right is the first
		###to be tested. So, vertex specifications are [ 7 6 5 4 3 2 1 0 ], [ 15 14 13 12 11 10 9 8 ], ...
		###"Less Than" is used since the loop incrementer will be adding '8' instead of '1'
		while j < CSK_MODEL_VERTEX_COUNT:
			###This value will be shifted once per loop
			loopshiftcount = 8
			###I screwed up some indexing plans, so this value is a crutch.. oh well.
			currentVnum = j
			bintest = 0b00000001
			binval = struct.unpack('<B', bitArray[morphBitArrayCount])[0]
			###The very last byte will often have less than 8 relavant bits. If the total model vert count minus the
			###current outer loop index is less than 8, there are only that number of bits left to test
			if (CSK_MODEL_VERTEX_COUNT-j) < loopshiftcount:
				loopshiftcount = CSK_MODEL_VERTEX_COUNT-j
			while loopshiftcount != 0:
				###get the current BASE VERTEX ARRAY index.. oi, this gets complicated. Time for some Weird Al. Perform This Way!
				###bitwise AND with the shifting mask. If true, this vertex has morphed.
				###IF IT MORPHED: Calculate the displacement and write out the new vertex position
				###IF NOT MORHPED: Write out the original vertex position
				if (bintest & binval):
					###remember that the vertex Z and the normal Z may need to be negated
					vertexString += "v " + str(format(CSKVerts[currentVnum][0]+VertexMorphTuple[morphVertexArrayCount][0], ' .8f')) + " " + str(format(CSKVerts[currentVnum][1]+VertexMorphTuple[morphVertexArrayCount][1], ' .8f')) + " " + str(format((CSKVerts[currentVnum][2]+VertexMorphTuple[morphVertexArrayCount][2])*NVZ, ' .8f')) + "\n"
					normalString += "vn " + str(format(CSKNormals[currentVnum][0]+VertexMorphTuple[morphVertexArrayCount][3], ' .8f')) + " " + str(format(CSKNormals[currentVnum][1]+VertexMorphTuple[morphVertexArrayCount][4], ' .8f')) + " " + str(format((CSKNormals[currentVnum][2]+VertexMorphTuple[morphVertexArrayCount][5])*NNZ, ' .8f')) + "\n"
					morphVertexArrayCount += 1
				else:
					vertexString += "v " + str(format(CSKVerts[currentVnum][0], ' .8f')) + " " + str(format(CSKVerts[currentVnum][1], ' .8f')) + " " + str(format((CSKVerts[currentVnum][2]*NVZ), ' .8f')) + "\n"
					normalString += "vn " + str(format(CSKNormals[currentVnum][0], ' .8f')) + " " + str(format(CSKNormals[currentVnum][1], ' .8f')) + " " + str(format((CSKNormals[currentVnum][2]*NNZ), ' .8f')) + "\n"
				bintest = bintest << 1
				currentVnum += 1
				loopshiftcount -= 1
			###Prepare to read the next byte in the bit array
			morphBitArrayCount += 1
			###We just read 8 vertices in the inner loop, not 1
			j += 8

################################################
#           Write to morph file, and close     #
################################################
#Realize that face information will never change
#the same faces always exist
		owfile.write(vertexString)
		owfile.write(normalString)
		owfile.write(UVString)
		###If faces are being extracted, write the face string calculated from the
		###base model
		if extractFaces:
			owfile.write(faceString)
		owfile.close()
		i += 1

	orfile.close()

	print " done"
################################################
################################################










################################################
#              Extract CCS Model               #
################################################
difParts = 0
currentNumParts = 0
currentParts = [ -1, -1, -1 ]
currentFaceIndex = 0
ccsV1 = 0
ccsV2 = 0
ccsV3 = 0
cskV1 = 0
cskV2 = 0
cskV3 = 0
tempV = 0

if modelExpChoice == 1:

	print "Extracting CCS model . . .",

	###initialize face used array.
	i = 0
	while i < CCS_FACE_COUNT:
		faceUsed[i] = 0
		i += 1


	###Prepare verticies
	i = 0
	while i < CCS_VERTEX_COUNT:
		###Remember, NVZ = inverted Z modifier. will be 1 if no inversion, -1 if inversion
		vertexString += "v " + str(format(CCSVerts[i][0], ' .8f')) + " " + str(format(CCSVerts[i][1], ' .8f')) + " " + str( format((CCSVerts[i][2]*NVZ), ' .8f') ) + "\n"
		i += 1
	vertexString += "\n"


	###Ready normals
	i = 0
	while i < CCS_VERTEX_COUNT:
		###NVZ = inverted Z modifier. will be 1 if no inversion, -1 if inversion
		normalString += "vn " + str(format(CCSNormals[i][0], ' .8f')) + " " + str(format(CCSNormals[i][1], ' .8f')) + " " + str( format((CCSNormals[i][2]*NNZ), ' .8f') ) + "\n"
		i += 1
	normalString += "\n"


	###Ready coordinates
	###Realize that the CCS model does not have any UV coordinates. Write out the CSK coordintes and associate
	###Them with the correct CCS vertices in the face list. Realize that this means that some verts (on seams)
	###will have more than one UV point when specified in different faces
	i = 0
	while i < CSK_MODEL_VERTEX_COUNT:
		if NUV:
			if CSKUV[i][0] > 0.5:
				UVString += "vt " + str( format(((1-CSKUV[i][0])*2), ' .8f') ) + " " + str( format((1 - CSKUV[i][1]), ' .8f') ) + "\n"
			else:
				UVString += "vt " + str( format(CSKUV[i][0]*2, ' .8f') ) + " " + str( format(1 - CSKUV[i][1], ' .8f') ) + "\n"
		else:
			UVString += "vt " + str(format(CSKUV[i][0], ' .8f')) + " " + str(format(CSKUV[i][1], ' .8f')) + "\n"
		i += 1
	UVString += "\n"





	###Prepare faces. Use the face group arrays in the CCS to create the body-part groups
	###Since this is the CCS model, the vertex indexes are the actual indexes. However, 
	###The UV points must be modified to use the CSK UV points by associating the CCS
	###verts with the CSK verts. Some CCS verts along UV seams will have multiple UV points
	i = 0
	while i < CCS_BODY_PART_COUNT:
		###j will be the number of faces in this part's array. +1 since the index is starting from
		###"1" to accomodate index scheme from CCSFaceGroupArrays
		j = CCSFaceGroupArrays[i][0] + 1
		k = 1
		while k < j:
			difParts = 0
			currentFaceIndex = CCSFaceGroupArrays[i][k]
			if faceUsed[currentFaceIndex] != 1:
				faceUsed[currentFaceIndex] = 1
				ccsV1 = CCSFaces[currentFaceIndex][0]
				ccsV2 = CCSFaces[currentFaceIndex][1]
				ccsV3 = CCSFaces[currentFaceIndex][2]
				cskV1 = CSKFaces[CCStoCSKFaces[currentFaceIndex]][0]
				cskV2 = CSKFaces[CCStoCSKFaces[currentFaceIndex]][1]
				cskV3 = CSKFaces[CCStoCSKFaces[currentFaceIndex]][2]

				###The face vertices between the CSK and the CCS will be the same and
				###in the same rotation direction (counter/clock-wise), but may not
				###actually be in the same order. For the UV indexes to be written
				###correctly, the orders must match. Find the correct order. Check
				###X,Y,Z of first CCS vert against the 2nd and 3rd CSK vert. If they
				###match, adjust the CSK vert order here accordingly. If the 1st
				###matches with the 1st, no adjustment since they're already in the
				###correct order.
				if (abs(CCSVerts[ccsV1][0]-CSKVerts[cskV2][0]) < 0.0002) and (abs(CCSVerts[ccsV1][1]-CSKVerts[cskV2][1]) < 0.0002) and (abs(CCSVerts[ccsV1][2]-CSKVerts[cskV2][2]) < 0.0002):
					tempV = cskV1
					cskV1 = cskV2
					cskV2 = cskV3
					cskV3 = tempV
				if (abs(CCSVerts[ccsV1][0]-CSKVerts[cskV3][0]) < 0.0002) and (abs(CCSVerts[ccsV1][1]-CSKVerts[cskV3][1]) < 0.0002) and (abs(CCSVerts[ccsV1][2]-CSKVerts[cskV3][2]) < 0.0002):
					tempV = cskV3
					cskV3 = cskV2
					cskV2 = cskV1
					cskV1 = tempV

				if currentNumParts != len(CCSFaceToPart[currentFaceIndex]):
					currentNumParts = len(CCSFaceToPart[currentFaceIndex])
					currentParts = CCSFaceToPart[currentFaceIndex]
					faceString += "g "
					m = 0
					while m < currentNumParts:
						faceString += "BodyPart_" + str(CCSFaceToPart[currentFaceIndex][m]) + " "
						m += 1
					faceString += "\n"
					faceString += "f " + str(ccsV1+1) + "/" + str( cskV1+1 ) + "/" + str(ccsV1+1) + " " + str(ccsV2+1) + "/" + str( cskV2+1 ) + "/" + str(ccsV2+1) + " " + str(ccsV3+1) + "/" + str( cskV3+1 ) + "/" + str(ccsV3+1) + "\n"
				else:
					m = 0
					while m < len(CCSFaceToPart[currentFaceIndex]):
						if CCSFaceToPart[currentFaceIndex][m] != currentParts[m]:
							difParts = 1
						m += 1
					if difParts:
						currentNumParts = len(CCSFaceToPart[currentFaceIndex])
						currentParts = CCSFaceToPart[currentFaceIndex]
						faceString += "g "
						m = 0
						while m < currentNumParts:
							faceString += "BodyPart_" + str(CCSFaceToPart[currentFaceIndex][m]) + " "
							m += 1
						faceString += "\n"
						faceString += "f " + str(ccsV1+1) + "/" + str( cskV1+1 ) + "/" + str(ccsV1+1) + " " + str(ccsV2+1) + "/" + str( cskV2+1 ) + "/" + str(ccsV2+1) + " " + str(ccsV3+1) + "/" + str( cskV3+1 ) + "/" + str(ccsV3+1) + "\n"
					else:
						faceString += "f " + str(ccsV1+1) + "/" + str( cskV1+1 ) + "/" + str(ccsV1+1) + " " + str(ccsV2+1) + "/" + str( cskV2+1 ) + "/" + str(ccsV2+1) + " " + str(ccsV3+1) + "/" + str( cskV3+1 ) + "/" + str(ccsV3+1) + "\n"

			k += 1
		i += 1


	###If the option to extract faces was selected


	owfile.write(vertexString)
	owfile.write(normalString)
	owfile.write(UVString)
	if extractFaces:
		owfile.write(faceString)

	owfile.close()

	print " done"

################################################
################################################










