#
# Black and White OBJ creature model Importer V0.4
# ------------------------------------------------
#
# 
# This script takes one or more .OBJ files and imports them as a BW2
# creature. All four of the creature model files must be in the same folder!
# .CCS .CSK .CBN .CHA
# Many parts of the files use constant offsets or header sizes. Hard-coded
# addresses and offsets in this code reflect that.
#
# USAGE: arguments should be OBJ files to be imported. At least one must be
# specified, the base model OBJ of the form "modelname.obj". The other arguments
# should be the morph models of the form "modelname-morphname.obj". The morhpname
# has a 32-character limit. The game really only recognizes a limited set of 
# names: thin, fat, strong, weak, young, old, good, evil
#
# You must have Python installed for this to work. Written around Python 2.6.
# I am not a programmer.. data structures and solutions will generally
# not be efficiently written, as well as not using functions
#
#
#
# This script takes a series of OBJ model files and imports them into
# Black and White 2 creature model files ( CBN CCS CHA CSK ). Currently, only two
# of the files are generated, .CCS and .CSK. All four of the
# creature model files must be in the same folder! They need not be the originals,
# but they must be 'correct' in format and data. Don't use files which are
# already having problems.
# The script will take files of the form "modelname.obj" as the base and
# "modelname-morphname.obj" as the model morphs (bwolf.obj, bwolf-thin.obj ..)
# The output will be as "modelname__EXPORTED_X" where 'X' is an incrementing
# number.. the script will not overwite an already existing file.
#
# This is NOT a 'well written' thing and probably will not be.. I'm not a programmer..
# Could use functions, but that takes work.. much easier to copy-paste code
#
#
#
#
# OBJ files can only contain mesh skin information (vertices, normals, UV).
# Bones, bone weights, and so forth cannot be transferred in OBJ files. Therefore,
# information like bone weights, bones, 'material' specifications, body-part
# groups, and so forth are simply copied from the CSK and CCS file.
#
# Even though many sections are copied from the CSK / CCS, this is written to actually read
# in those sections 'correctly', unpacking the data into appropriate structures. This
# was done in anticipation of being able to eventually manipulate the data instead
# of having to only copy it. For now, it is useless work.
#
# It seems that OBJ formats are not standard. Some programs assign one normal/Uv per vertex,
# others assign the normal/UV per face. Both are technically correct, but create logical
# problems. Most 3D programs will expect one or the other and may import with some degree
# of error if the wrong form is used. I am of the opinion that the per-vertex specification is
# probably 'more' correct. Black and White 2 (and I think all DirectX/OpenGL) uses vector/UV
# information per-vertex.
# OBJ files must be interpreted and dealt with accordingly. This is made easier for this
# script since it expects vertex order to have been maintained.
# Blender's OBJ export is one program that I know exports normals per-face. The correct way to
# handle this is to take all the faces a vertex belongs to and average the normal vector
# of each face to give the vertex its own normal.
# Keep in mind that this script expects the OBJ model mesh to already be in a correct format
# for the game. There MUST be split seams along the UVs (multiple vertices for UV
# discontinuities), normals must exist (zbrush doesn't always export normals by default,
# for instance)
# Finally, note that the B&W2 game expects normals at locations with multiple vertices to
# be the same for all of the vertices at that point (or, at least that's what the retail
# models do). Normals for multiple vertices are averaged and normalized to create one normal
# that all the vertices at that point use.
#
#
#	V0.4
#	 - Significant changes. Portions of the script were rewritten or
#	   reordered to accomodate new knowledge of the CCS file.
#	 - CCS file is now generated as well as the CSK file.
#	V0.3
#	 - Improved tangent calculation. Added check to avoid unlikely but somehow
#	   existent divide-by-zeros. Fixed mistake in bitangent calculation.
#	   Incidently, update notes on CSK to reflect new info.
#	V0.2
#	 - Significant cleanup. Switched to reading whole files into
#       memory rather than reading small chunks off the disk
#	 - Added alot of commentary
#	TO DO:
#	- Write an un-weld ability? May be unneccesarily difficult
#
#
# string, string, string
# The first index is the base model
# the base model's morphname is the name of the modelfile. The other's are just the name
# of the morph, like "fat", "thin", "weak", ..
# Files =
#    [ ( path-to-the-file, the-filename-with-extension, name-of-the-morph ), ... ]
#
# (int, string) , [ (array of int) ] , [int, int, int] , [int, int] ]
# facearrayMin/Max are the min and max VALUES of the VERTEX REFERENCES in the face array. These are not
# actually read from the material/body-part arrays, but determined by the later reading of the face array
# CSKMats =
#   [ (array-count, Name) , [(partarray), [num-parts,index,count , [faceArrayMin, faceArrayMax] ,  ..repeat..]
#                            (partarray)   num-parts,index,count   [faceArrayMin, faceArrayMax]
#                            (partarray)]  num-parts,index,count]  [faceArrayMin, faceArrayMax]
#
#
#
# (float, float)
# CSKUV = 
#   [ (U, V), (U, V), ... ]
#
#
# (int, float, int, float, int, float)
# CSKBoneWeights =
#   [ ( bone1, weight1, bone2, weight2 ,bone3, weight3 ) , ... ]
#
#
# (int, int, int)
# CSKFaces / CCSFaces / OBJFaces = 
#   [ (v1, v2, v3), (v1, v2, v3), ... ]
#
#
# [ [int, int, .. , int] ]
# Each vertex can belong to varying numbers of faces. len(CSKVertexFaceindex[X]) gives
# the number of faces vertex number X belongs to.
# CSKVertexFaceIndex = 
#   [ [ face1, face2, .. faceX ] , [ face1, face2, .. faceX ] , .... ]
#
#
# (float, float, float)
# CSKVerts / CCSVerts / OBJVertex = 
#   [ ( X , Y , Z ) , ... ] - floats, up to the number of vertices in the model
#
#
# (float, float, float)
# CSKNormals / CCSNormals / OBJNormals / OBJNormalizedNormals = 
#   [ (NX, NY, NZ), ... ] - directions of normal vectors for vertices
#
#
# (float, float, float)
# OBJTangents =
#   [ (TanX, TanY, TanZ), ... ]
#
#
# (float, float, float)
# OBJBitangents =
#   [ (TanX, TanY, TanZ), ... ]
#
#
# All floats
# CSKTANBI = 
#   [ (Tan1x, Tan1y, Tan1z, Bitan1x, Bitan1y, Bitan1z, Tan2x, Tan2y, Tan2z, Bitan2x,
#      Bitan2y, Bitan2z), ... ]
#
#
# int, int, int, int
# CSKEdgeConn = 
#   [ ( EdgeFaceIndex, CCSVertIndex, OppositeEdge, NextEdge), ... ]
#
#
# all ints
# CSKConn256 = 
#   [ ( Edge1, Edge2, Edge3, material, part-array, CSKVert1, CSKVert2, CSKVert3), ... ]
#
#
# All ints
# CCSIKChains = 
#   [ [ part1, part2, part3, ...], [ part1, part2, part3, ...], ... ]
#
#
# all floats
# CCSBoundingBox = 
#   [ (X1, Y1, Z1, X2, Y2, Z2), ... ]
#
#
# all ints
# CCSFaceGroupArrays = 
#   [ [partCount, part1, part2, ...], [partCount, part1, part2, ... ], ... ]
#
#
#
# TO DO:
#	- Why do eyes have purposefully non-normal normals?
#######################################################################################



import struct
import math
import sys
import os

################################################
#           Variables Used                     #
################################################

#Values considered as 'constants' for the model
#either read in or calculated.
CSK_NUM_MATS = 0
CSK_BODY_PART_MAX = 0
CSK_FACE_ARRAY_ENTRY_COUNT = 0
CSK_VERT_META_ARRAY_SUBARRAY_SIZE = 0
CSK_VERT_META_ARRAY_SIZE = 0
CSK_BASE_VERTEX_ARRAY_SUB_ARRAY_SIZE = 0
CSK_BASE_VERTEX_ARRAY_SIZE = 0
CSK_MODEL_VERTEX_COUNT = 0
CSK_MORPH_COUNT = 0
CSK_FACE_COUNT = 0
CSK_EDGE_COUNT = 0


CCS_IK_ARRAY_NUM = 0
CCS_IK_ARRAY_ENTRY_NUM = 0
CCS_VERTEX_COUNT = 0
CCS_FACE_COUNT = 0
CCS_BODY_PART_COUNT = 0


#"constants" changed within various sections
MORPH_VERTEX_ARRAY_SUB_ARRAY_SIZE = 0
MORPH_VERTEX_ARRAY_SIZE = 0
NUM_PART_ARRAYS = 0


#CSK data structure addresses
#----------------------------
#Face array address
CSKfaceArrayLoc = 0
#UV/Meta array address
CSKUVArrayLoc = 0
#Vertex array index
CSKVertexArrayLoc = 0
#Where the morph sections begin
CSKMorphSectionBegin = 0


#New CCS data structure addresses
#----------------------------
#File size number location
NewCCSwriteFileSizeLoc = 0


#New CSK data structure addresses
#----------------------------
#Address of new CSK file's file-size definition
NewCSKwriteFileSectionSizeLoc = 0
#Address of new CSK file's face array size definition
NewCSKfaceArrayCountLoc = 0



#Used to push current position for quick jumps back and forth
tempLoc = 0



#Various indecies
#----------------
i = 0
j = 0
k = 0
m = 0
h = 0
u = 0


#scratch space
scratchvalue = 0

#holds temporary values
valuebuffer = ''



#Hold entire contents of original CSK in memory
CSKFilebuffer = ''
#Holds a current address into the CSKFilebuffer used for reading
#variable length portion of the file
CSKFilebufferPT = 0


#Hold entire contents of CCS in memory
CCSFilebuffer = ''
#Holds a current address into the CCSFilebuffer used for reading
#variable length portion of the file
CCSFilebufferPT = 0


#Holds entire OBJ files in memory for reading
OBJFileBuffer = 0



#Filename vars
#-------------
#filename of the model file given
filename = ''
#filename of the model file given without extension
filenamenoext = ''
#extension only of the model filename
filenameext = ''
#full path to the model file, plus the filename, but no extension
pathtomodelnoext = ''
#The directory where the files are with a '/' at the end
filedir = ''
#File handle of file currently being read
orfile = ''
#File handles of file currently being written
owfileCSK = ''
owfileCCS = ''

#Holds file argument data
files = []



#Option related Vars
#-------------------
userInput =''
#(UN-)Negate Vertex Z
NVZ = 1
#(UN-)Negate UV's (misnomer.. not really negating, just re-arranging)
NUV = 0
#(UN-)Negate Normal Z
NNZ = 1
#Should the model be (UN-)welded?
#weld = 0



#CSK
#------------

#holds verbatim the 384-byte section, who's purpose I've never really been able to discover
CSK384byteArray = ''

#holds the verbatim copy of the face array. If ever there is the ability to generate
#a new face array (to change the mesh), this won't be needed.
CSKFaceArrayCopy = ''


#Holds data describing the original CSK's material sections, *plus*
#the min and max vertex index numbers of the associated section of
#the face array (not present in Creature2OBJ)
CSKMats = []

#Holds the vertices read from the original CSK model
CSKVerts = []

#Holds normal data for each vertex read from the original CSK file. The
#indexes match the index from the vertex array
CSKNormals = []

#Holds tangent and bitangent data for each vertex read from the original
#CSK file. Indexes match the index from the vertex array
CSKTANBI = [] 

#Holds array of all UV points read in from the original CSK
CSKUV = []

#Holds the bone association and the weight for each association
#There should be one index per vertex, in the same order as the
#verticies.
CSKBoneWeights = []

#Holds triplets that describe faces from the original CSK model.
#Each entry in a triplet is an index of a vertex as read in
#the CSKverts array.
CSKFaces = []


#Holds a 'reverse' face lookup for vertices. Given a vertex number as an
#index, that index will hold an array of the faces (face indexes) which
#that vertex belongs to. Quickly find all faces using a particular vertex
CSKVertexFaceIndex = []


#Holds the unpacked data from the section called
#EdgeConnectivity
CSKEdgeConn = []

#Holds the unpacked data from the section called
#Connectivity256Const
CSKConn256 = []

#Using the edge sections in the CSK, an association of the CSK
#verticies to CCS vertices. Note that this is NOT a directly
#reversible process. A single CCS vertex will equate to
#multiple CSK verticies.
#INDEX = CSK model vertex index
#VALUE = CCS model vertex index
CSKtoCCSVerts = []





#CCS
#------------


#Holds a copy of the eye to face association array. This array defines
#which model faces belong to the rotatable part of the eyes. Since the
#mesh cannot be modified, this cannot change, so just copy it.
CCSEyeFacesArrayCopy = ''


#Holds the info from what is currently guessed to be 'IK chains'
CCSIKChains = []

#Holds the X,Y,Z tuples for the vertices from the CCS model
CCSVerts = []

#Holds the body part association for each vertex in the CCS model.
#The index corresponds to a vertex and the value specifies the
#body-part number the vertex belongs to
#INDEX = CCS model vertex index
#VALUE = body-part number
CCSVertPartArray = []

#Holds the X,Y,Z values of normals for each vertex in the CCS model.
#The index corresponds to a vertex
CCSNormals = []

#Holds the face info for the CCS. Each index is a face, the value is
#a tuple holding indexes of the vertices that make up the face
CCSFaces = []

#Holds the bounding boxes for the CCS model. The index corresponds to
#a body-part number. The values are two X,Y,Z coordinates that define
#two points. These two points define a 3D box that should encompass
#the body-part
CCSBoundingBox = []

#Array of arrays that holds the face-to-bodypart association for the
#CCS mode. The index corresponds to a body part. The sub-array is a
#list of face index numbers that indicate that the face belongs to that
#body part. The FIRST value in the sub-array is a count of how many
#faces are in the sub-array. A face can belong to more than on array.
CCSFaceGroupArrays = []

#Holds a calculated "reverse-lookup" for the part arrays contained
#in the CCS. Each index corresponds to a face and the values indicate
#which body parts the face belongs to. Each face can belong to up to 3
#different parts
#INDEX = CCS model face index
#VALUES = body part numbers (up to three)
CCSFaceToPart = []

#Holds a partial association of CCS vertex index to CCS vertex. It is
#'partial' because it is not quite accurate. This is used to create the
#new CCS vertex and normal data, where only one duplicate vertex is
#needed. A single CCS vert index will actually have multiple CSK verts
#associated with it.
CCStoCSKVerts_partial = []






#OBJ
#------------

#Temporary arrays used to read in values from the OBJ files.
#file contents are kept here until they are correctly processed
#into actual arrays
OBJNormalTemp = []
OBJUVTemp = []
OBJNormalcollector = []
OBJUVcollector = []

#Holds the verticies for the base OBJ model
OBJVertex = []

#Holds the *Correctly Calculated* UVs for the base OBJ model,
#not the values read in from the OBJ
OBJUV = []

#Holds the averaged normals for the base OBJ model. Any single
#vertex (not a single point with multiple vertecies yet!) that
#had multiple normals specified in the OBJ will have a single
#averaged normal contained here
OBJNormals = []

#Holds the final, averaged and normalized normals for each vertex
OBJNormalizedNormals = []

#CURRENTLY UNUSED, and INCORRECT. This would be used if the face array
#were to be regenerated. Also, every even-numbered entry should be re-
#arranged as (v2, v1, v3).
OBJFaces = []

#Each index corresponds to a vertex. The values
#the index of any other vertex that is a duplicate of this vertex
OBJVertexDuplicates = []

#Holds the recalculated tangents for each vertex read from the OBJ
OBJTangents = []

#Holds the recalculated bitangents for each vertex read from the OBJ
OBJBitangents = []






#Holds temporarily unpacked arrays while data is processed
tempArray = []
tempArray2 = []


#Holds array of "full" edges. index will be from 0 to the number of faces
#in the model *3 (There are three edges in each face). Each index holds
#two vertex indicies that indicate which vertices form the edge.
#The INDEX number of an entry is edge number of that edge
VertexEdgeArray = []

#Indicates which two edges in full VertexEdgeArray are twins. Since an edge
#will usually be used as an edge for two faces, this means that an edge is
#usually specified twice in the VertexEdgeArray array. The exception are
#edges on the UV seams; these don't have a twin because the duplicate
#vertecies at those seam points break the model continuity. These edges
#are indicated with an integer value of "-1" (as the CSK file actually does
DupEdgeArray = []





###########################################
#        Arguments                        #
###########################################
morphname = ''

if len(sys.argv) < 2:
	print 
	print "USAGE: OBJ2Creature.py modelname.obj [ modelname-morphname.obj ]"
	print "multiple morph obj files may be specified. The specified morphs"
	print "will be inserted into the exported CSK."
	print "\"morphname\" should be any one of:"
	print "young, old, weak, strong, fat, thin, evil, good"
	print "but other names can be used. The game will not recognize other names unless"
	print "the game been modified"
	print
	print "ex: OBJ2Creature.py bwolf.obj bwolf-thin.obj bwolf-fat.obj bwolf-evil.obj"
	print
	print "All model files must be present in the same folder as the base model obj"
	print " ( CCS CBN CHA CSK ) - some values will be copied from these original files"
	print
	raise SystemExit


files = []
###The first argument is the script name, skip "0"
i = 1
###The number of files other than the base morph are the number of morphs
CSK_MORPH_COUNT = len(sys.argv) - 2
while i < len(sys.argv):
	morphname = ''
	filename = os.path.basename(sys.argv[i])
	filedir = os.path.dirname(sys.argv[i])
	if filedir != "":
		filedir += "/"
	filenamenoext, filenameext = os.path.splitext(filename)
	pathtomodelnoext = filedir+filenamenoext
	j = 0
	###The base model name won't have a dash..
	if i != 1:
		while j < (len(filenamenoext)) and filenamenoext[j] != "-":
			j += 1
		###add one more to proceed past the dash
		j += 1
	morphname = filenamenoext[j:len(filenamenoext):1]
	if len(morphname) > 32:
		print
		print "ERROR:Morph names must be 32 characters or less"
		print "For: " + filename
		print
		raise SystemExit
	if len(morphname) == 0:
		print
		print "ERROR:Filename of Morph does not match"
		print "base model filename."
		print "For: " + filename
		print
		raise SystemExit
	if not os.path.isfile(filedir+filename):
		print
		print "ERROR:OBJ file "
		print filename
		print "does not exist"
		print
		raise SystemExit
	files.append( (filedir, filename, morphname) )
	i += 1

###Check to see that the other model files exist
if not os.path.isfile(files[0][0] + files[0][2] + ".ccs"):
	print
	print "ERROR:CCS file does not exist"
	print
	raise SystemExit
if not os.path.isfile(files[0][0] + files[0][2] + ".cha"):
	print
	print "ERROR:CHA file does not exist"
	print
	raise SystemExit
if not os.path.isfile(files[0][0] + files[0][2] + ".csk"):
	print
	print "ERROR:CSK file does not exist"
	print
	raise SystemExit
if not os.path.isfile(files[0][0] + files[0][2] + ".cbn"):
	print
	print "ERROR:CBN file does not exist"
	print
	raise SystemExit
################################################
################################################




###########################################
#     Open new files for writing          #
###########################################
#Don't overwrite files when outputting.. check for export name and
#increment the number up to 99
#Note that this file is being open for writing BYTES, not strings.

###CSK
owfileCSK = ''
i = 1
while not owfileCSK and i < 100:
	if not os.path.isfile(files[0][0] + files[0][2] + "__EXPORTED_" + str(i) + ".csk"):
		try:
			owfileCSK = open(files[0][0] + files[0][2] + "__EXPORTED_" + str(i) + ".csk", 'wb')
		except:
			print
			print "ERROR:Cannot open file for writing."
			print "Disk Full? Directory read-only?"
			print
			raise SystemExit
	else:
		i += 1
if i == 100:
	print
	print "ERROR:Cannot open file for writing"
	print "There can only be 99 Exported model"
	print "file sets. Do you really have 99?"
	print "Delete some of them and try again"
	print
	raise SystemExit

###CCS
owfileCCS = ''
i = 1
while not owfileCCS and i < 100:
	if not os.path.isfile(files[0][0] + files[0][2] + "__EXPORTED_" + str(i) + ".ccs"):
		try:
			owfileCCS = open(files[0][0] + files[0][2] + "__EXPORTED_" + str(i) + ".ccs", 'wb')
		except:
			print
			print "ERROR:Cannot open file for writing."
			print "Disk Full? Directory read-only?"
			print
			raise SystemExit
	else:
		i += 1
if i == 100:
	print
	print "ERROR:Cannot open file for writing"
	print "There can only be 99 Exported model"
	print "file sets. Do you really have 99?"
	print "Delete some of them and try again"
	print
	raise SystemExit
################################################
################################################




################################################
#           Get Extraction Options             #
################################################
print
print
print "CTRL-C at any point to break this script."
print
print "Remember that this scripts can only import the"
print "CSK version of the model. If the OBJ's are from"
print "the CCS model or if the mesh has been altered,"
print "the script will fail"
print
print
print "Should Vertex Z be (re)-inverted?"
print "  1) Y"
print "  2) N"
print
while userInput != "1" and userInput != "2":
	userInput = raw_input('--> ')
	if userInput != "1" and userInput != "2":
		print "Input is not correct"
if userInput == "1":
	NVZ = -1

userInput = ''
print
print
print "Should Normal vector Z be (re)-inverted?"
print "  1) Y"
print "  2) N"
print
while userInput != "1" and userInput != "2":
	userInput = raw_input('--> ')
	if userInput != "1" and userInput != "2":
		print "Input is not correct"
if userInput == "1":
	NNZ = -1

userInput = ''
print
print
print "Should UV's be arranged to match the structure"
print "needed by the game?"
print "If the UV's are in the format matching the .dds"
print "texture files, this needs to be done. If you"
print "notice the textures are off in the game, try"
print "rotating/flipping the UV's and re-importing"
print "  1) Y"
print "  2) N"
print
while userInput != "1" and userInput != "2":
	userInput = raw_input('--> ')
	if userInput != "1" and userInput != "2":
		print "Input is not correct"
if userInput == "1":
	NUV = 1

print
print "----------------------------"
print
################################################
################################################





################################################
#           Read CSK file and get header       #
################################################
###Try to open the CSK
try:
	orfile = open(files[0][0] + files[0][2] + ".csk", 'rb')
except:
	print
	print "ERROR:Could not open original CSK file for reading."
	print "Why? File seems to exist. Persmission problems?"
	print
	owfileCSK.close()
	owfileCCS.close()
	raise SystemExit


print "Reading CSK . . .",

orfile.seek(0)
CSKFilebuffer = orfile.read()

###Skip the file header info
CSKFilebufferPT = 52

###Get the maximum body part count
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
CSKFilebufferPT += 4
CSK_BODY_PART_MAX = struct.unpack_from('<i', valueBuffer)[0]


###If the first section is not 'radeon', Skip it. In that case,
###Gets the size of the first section, then jumps ahead that
###number of bytes.
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+32]
CSKFilebufferPT += 32
if valueBuffer[0] != "r" and valueBuffer[0] != "R":
	valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
	CSKFilebufferPT += 4
	scratchvalue = struct.unpack_from('<i', valueBuffer)[0]
	###Skip the section
	CSKFilebufferPT += scratchvalue
	###Skip the second section's header
	CSKFilebufferPT += 36
else:
	###If the section IS 'radeon', skip the section-size value
	CSKFilebufferPT += 4
################################################
################################################





################################################
#           Read CSK material sections         #
################################################
arraysRead = 0
matName = ''


###The Mat array data structure is documented at the top of this file.
###NOTE that this is NOT the same structure as the export Mat array
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
CSKFilebufferPT += 4
CSK_NUM_MATS = struct.unpack_from('<i', valueBuffer)[0]

i = 0
while i != CSK_NUM_MATS:
	CSKFilebufferPT += 4
	matName = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+128]
	CSKFilebufferPT += 128
	if i == 0:
		CSK384byteArray = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+384]
		CSKFilebufferPT += 384
	else:
		CSKFilebufferPT += 384
	valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
	CSKFilebufferPT += 4
	NUM_PART_ARRAYS = struct.unpack_from('<i', valueBuffer)[0]
	CSKMats.append( ( int(NUM_PART_ARRAYS), matName ) )
	CSKMats.append([])
	CSKMats.append([])
	CSKMats.append([])
	j = 0
	while j != NUM_PART_ARRAYS:
		valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
		CSKFilebufferPT += 4
		PART_ARRAY_COUNT = struct.unpack_from('<i', valueBuffer)[0]
		CSKMats[(i*4)+2].append(PART_ARRAY_COUNT)
		### Read the array and the two count numbers
		valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+(PART_ARRAY_COUNT * 4)]
		CSKFilebufferPT += (PART_ARRAY_COUNT * 4)
		CSKMats[(i*4)+1].append(struct.unpack_from('<'+str(PART_ARRAY_COUNT)+'i', valueBuffer))
		valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+8]
		CSKFilebufferPT += 8
		CSKMats[(i*4)+2].append(struct.unpack_from('<i', valueBuffer)[0])
		CSKMats[(i*4)+2].append(struct.unpack_from('<2i', valueBuffer)[1])
		arraysRead += 1
		j += 1
	i += 1
################################################
################################################




################################################
#           Skip CSK Face Array                #
################################################
###Now at the beginning of the Face array.
###For convenience, this should be done later.
###Keep the address
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
CSKFilebufferPT += 4
CSK_FACE_ARRAY_ENTRY_COUNT = struct.unpack_from('<i', valueBuffer)[0]
CSKfaceArrayLoc = CSKFilebufferPT
###The entries in this array are words (2 bytes)
CSKFilebufferPT += (CSK_FACE_ARRAY_ENTRY_COUNT * 2)
################################################
################################################




################################################
#           Skip CSK UV/Meta array             #
################################################
###Now at the beginning of the UV arrays. These
###hold UV data and bone associations/weight for
###each vertex. Also skip it for now.
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
CSKFilebufferPT += 4
CSK_VERT_META_ARRAY_SUBARRAY_SIZE = struct.unpack_from('<i', valueBuffer)[0]
if CSK_VERT_META_ARRAY_SUBARRAY_SIZE != 28:
	print
	print "ERROR: UV array in unknown format"
	print "Each sub-array should be 28-bytes"
	print
	orfile.close()
	owfileCSK.close()
	owfileCCS.close()
	raise SystemExit
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
CSKFilebufferPT += 4
CSK_VERT_META_ARRAY_SIZE = struct.unpack_from('<i', valueBuffer)[0]
CSKUVArrayLoc = CSKFilebufferPT
CSKFilebufferPT += CSK_VERT_META_ARRAY_SIZE
################################################
################################################





################################################
#           Read CSK vertex array              #
################################################
###The vertex array contains (X,Y,Z), Normal,
###tangent, and bitangent info per vertex. There
###Are two tan/bitan pairs per vertex (12 numbers)
###but I don't know what effect the second pair has


###Begin extracting the vertices
CSKFilebufferPT += 32
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
CSKFilebufferPT += 4
CSK_BASE_VERTEX_ARRAY_SUB_ARRAY_SIZE = struct.unpack_from('<i', valueBuffer)[0]
if CSK_BASE_VERTEX_ARRAY_SUB_ARRAY_SIZE != 72:
	print
	print "ERROR:Vertex array in unknown format"
	print "Each sub-array should be 72-bytes"
	print
	orfile.close()
	owfileCSK.close()
	owfileCCS.close()
	raise SystemExit
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
CSKFilebufferPT += 4
CSK_BASE_VERTEX_ARRAY_SIZE = struct.unpack_from('<i', valueBuffer)[0]
CSK_MODEL_VERTEX_COUNT = CSK_BASE_VERTEX_ARRAY_SIZE/CSK_BASE_VERTEX_ARRAY_SUB_ARRAY_SIZE

###Remember the position of the vertex array
CSKVertexArrayLoc = CSKFilebufferPT

valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+CSK_BASE_VERTEX_ARRAY_SIZE]
CSKFilebufferPT += CSK_BASE_VERTEX_ARRAY_SIZE

j = CSK_BASE_VERTEX_ARRAY_SIZE
i = 0
while i < j:
	CSKVerts.append(struct.unpack_from('<3f', valueBuffer[i:i+12:1]))
	CSKNormals.append(struct.unpack_from('<3f', valueBuffer[i+12:i+24:1]))
	CSKTANBI.append(struct.unpack_from('<12f', valueBuffer[i+24:i+72:1]))
	i += 72



###Save the morph section location. If the CSK model is
###being exported and the morphs are going to be extracted,
###it will be necessary to go back to this address
CSKMorphSectionBegin = CSKFilebufferPT


################################################
# Reconstruct faces from CSK face index array  #
################################################
#Holds the entire face array as an unpacked, parsable data structure
#Face array read first here, since the min/max vert index
#numbers are needed in the UV array reading.
faceTuple = ''
numArrays = 0
endArrayEntryNum = 0
minVertexNum = 0
maxVertexNum = 0


CSKFilebufferPT = CSKfaceArrayLoc

###Each entry is a word, 2-bytes
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+(CSK_FACE_ARRAY_ENTRY_COUNT * 2)]
CSKFilebufferPT += CSK_FACE_ARRAY_ENTRY_COUNT * 2

###While there is no ability to change the mesh, keep a copy of the face array
###to write into the new CSK file. This must be done since 'reconstructing' the
###array will lose data (the 'degenerative' part of the array).
CSKFaceArrayCopy = valueBuffer


###create an array of *words* of the entire face array
faceTuple = struct.unpack_from('<' + str(CSK_FACE_ARRAY_ENTRY_COUNT) + 'h', valueBuffer)


i = 0
while i != CSK_NUM_MATS:
	###numArrays will be the number of part arrays in this Material
	numArrays = CSKMats[i*4][0]
	j = 0
	while j != numArrays:
		###k will be the beginning index of this part-array's listings in the face array
		k = CSKMats[(i*4)+2][(j*3)+1]
		###endArrayEntryNum will be the last index number in the face array for this
		###part-array
		endArrayEntryNum = CSKMats[(i*4)+2][(j*3)+2] + k
		###Faces will be triplets, (index-1, index, index+1). For index-1 and index+1 to
		###exist correctly, the reading must start at k+1 and end at endArrayEntryNum-1
		k = k + 1
		endArrayEntryNum = endArrayEntryNum - 1
		h = 1
		###Set min/max vertex index to the first vertex. k-1 since k
		###actually starts on 2nd vertex entry
		minVertexNum = faceTuple[k-1]
		maxVertexNum = faceTuple[k-1]
		while k != endArrayEntryNum:
			if faceTuple[k] < minVertexNum:
				minVertexNum = faceTuple[k]
			if faceTuple[k] > maxVertexNum:
				maxVertexNum = faceTuple[k]
			###As long as any of the numbers in the triplet are not the same...
			if faceTuple[k] != faceTuple[k-1] and faceTuple[k] != faceTuple[k+1] and faceTuple[k-1] != faceTuple[k+1]:
				if h:
					CSKFaces.append( ( faceTuple[k-1] , faceTuple[k] , faceTuple[k+1] ) )
				else:
					CSKFaces.append( ( faceTuple[k] , faceTuple[k-1] , faceTuple[k+1] ) )
			h = not h
			k += 1
		###Test last vertex. note that K was incremented one last time
		###in the previous loop so 'k' is now the last vertex.
		if faceTuple[k] < minVertexNum:
			minVertexNum = faceTuple[k]
		if faceTuple[k] > maxVertexNum:
			maxVertexNum = faceTuple[k]
		CSKMats[(i*4)+3].append( ( int(minVertexNum) , int(maxVertexNum) ) )
		j += 1
	i += 1

CSK_FACE_COUNT = len(CSKFaces)
################################################
################################################






################################################
#           Read CSK UV/Meta array             #
################################################
#The UV array contains UV data and bone
#association/weights per vertex.
#The boneweight numbers read in are the 'encoded' numbers
#rather than the actual body-part numbers. If the model
#is ever to allow mesh modifications, it would probably
#be better to calculate the ACTUAL body-part number
#to go in the array. There is a code example below.



###Go back and get the UV texture coordinate array
CSKFilebufferPT = CSKUVArrayLoc
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+CSK_VERT_META_ARRAY_SIZE]
CSKFilebufferPT += CSK_VERT_META_ARRAY_SIZE

j = CSK_VERT_META_ARRAY_SIZE
i = 0
while i != j:
	CSKUV.append(struct.unpack_from('<2f', valueBuffer[i:i+8:1]))
	CSKBoneWeights.append( ( struct.unpack('<h', valueBuffer[i+8:i+10])[0], struct.unpack('<f', valueBuffer[i+16:i+20])[0], struct.unpack('<h', valueBuffer[i+10:i+12])[0], struct.unpack('<f', valueBuffer[i+20:i+24])[0], struct.unpack('<h', valueBuffer[i+12:i+14])[0], struct.unpack('<f', valueBuffer[i+24:i+28])[0] ) )
	i += 28


###If the actual body-part numbers are ever needed in the bone weight array
###the following code will do it.
#i = 0
#while i < CSK_NUM_MATS:
#	j = 0
#	###j counts through all of the part arrays
#	while j < CSKMats[(i*4)][0]:
#		###k will be the min vertex number for this part array
#		###m will be the max vertex number
#		###test is <= since the last vertex should actually be used
#		k = CSKMats[(i*4)+3][j][0]
#		m = CSKMats[(i*4)+3][j][1]
#		while k <= m:
#			###The first value will be the ACTUAL body-part number instead of the part number
#			###from the 'part-array' group divided by three.. 
#			CSKBoneWeights[k][0] = CSKMats[(i*4)+1][j][CSKBoneWeights[k][0]/3]
#			CSKBoneWeights[k][2] = CSKMats[(i*4)+1][j][CSKBoneWeights[k][1]/3]
#			CSKBoneWeights[k][4] = CSKMats[(i*4)+1][j][CSKBoneWeights[k][2]/3]
#			k += 1
#		j += 1
#	i += 1



################################################
#      Create Vertex->Face lookup array        #
################################################
#Additionally, create a face index array.. this is an array who's index
#corresponds to each vertex and who's values indicate which faces (as
#indexed in the face array) include this vertex
#Used to quickly determine which faces contain any particular vertex
i = 0
while i < CSK_MODEL_VERTEX_COUNT:
	CSKVertexFaceIndex.append([])
	i += 1
i = 0
while i < CSK_FACE_COUNT:
	CSKVertexFaceIndex[CSKFaces[i][0]].append(int(i))
	CSKVertexFaceIndex[CSKFaces[i][1]].append(int(i))
	CSKVertexFaceIndex[CSKFaces[i][2]].append(int(i))
	i += 1



################################################
#          Read CSK edge info                  #
################################################
edgeArraySize = 0

###The morph section of the CSK will have to be skipped first.
###It will take some effort to simply find the beinning of the
###edge arrays..
###First, go back to the beinning of the morph secion...
CSKFilebufferPT = CSKMorphSectionBegin
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
CSKFilebufferPT += 4
CSK_MORPH_COUNT = struct.unpack_from('<i', valueBuffer)[0]

i = 0
while i < CSK_MORPH_COUNT:
	###morph name
	CSKFilebufferPT += 32
	###Skip two counts, get the sub-array size
	valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+8]
	CSKFilebufferPT += 8
	MORPH_VERTEX_ARRAY_SUB_ARRAY_SIZE = struct.unpack_from('<i', valueBuffer)[0]

	###Skip bitwise array. 1 byte per vertex
	CSKFilebufferPT += CSK_MODEL_VERTEX_COUNT

	###Get morph array size
	valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
	CSKFilebufferPT += 4
	MORPH_VERTEX_ARRAY_SIZE = struct.unpack_from('<i', valueBuffer)[0]

	###Skip morph array
	CSKFilebufferPT += MORPH_VERTEX_ARRAY_SIZE

	i += 1

###Should now be at the start of the edge sections
###Skip header and size dword
CSKFilebufferPT += 36

###Get edge count 
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
CSKFilebufferPT += 4
CSK_EDGE_COUNT = struct.unpack_from('<i', valueBuffer)[0]

###Each edge subarray is 16 bytes (4 dwords)
edgeArraySize = CSK_EDGE_COUNT*16

valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+edgeArraySize]
CSKFilebufferPT += edgeArraySize


i = 0
while i < edgeArraySize:
	CSKEdgeConn.append( struct.unpack_from('<4i', valueBuffer[i:i+16]) )
	i += 16

###Now, skip the '96' section
CSKFilebufferPT += 32
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
CSKFilebufferPT += 4
edgeArraySize = struct.unpack('<i', valueBuffer)[0]
CSKFilebufferPT += edgeArraySize

###Get the '256' section
CSKFilebufferPT += 32
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+4]
CSKFilebufferPT += 4
edgeArraySize = struct.unpack('<i', valueBuffer)[0]
###Skip the face count.. already know it
CSKFilebufferPT += 4
###The edgecount array size included the face count number.. don't want
###this in the count, so sub 4.
edgeArraySize = edgeArraySize - 4

###Each sub-array is 32 bytes long (8 ints)
valueBuffer = CSKFilebuffer[CSKFilebufferPT : CSKFilebufferPT+edgeArraySize]
CSKFilebufferPT += edgeArraySize

i = 0
while i < edgeArraySize:
	CSKConn256.append( struct.unpack('<8i', valueBuffer[i:i+32]) )
	i += 32

valueBuffer = ''

################################################
################################################



orfile.close()

print " done"

################################################
################################################
################################################
################################################






################################################
#               Read CCS file                  #
################################################
###Try to open the CCS
try:
	orfile = open(files[0][0] + files[0][2] + ".ccs", 'rb')
except:
	print
	print "ERROR:Could not open original CCS file for reading."
	print "Why? File seems to exist. Persmission problems?"
	print
	owfileCSK.close()
	owfileCCS.close()
	raise SystemExit

print "Reading CCS . . .",

orfile.seek(0)
CCSFilebuffer = orfile.read()
CCSFilebufferPT = 0



###Skip CCS Header
CCSFilebufferPT += 44


################################################
#           Read CCS IK                        #
################################################

###Skip the IK entry count number
CCSFilebufferPT += 4


valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+4]
CCSFilebufferPT += 4
CCS_IK_ARRAY_NUM = struct.unpack('<i', valueBuffer)[0]



valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+4]
CCSFilebufferPT += 4
CCS_IK_ARRAY_ENTRY_NUM = struct.unpack('<i', valueBuffer)[0]


i = 0
while i < CCS_IK_ARRAY_NUM:
	j = 0
	CCSIKChains.append([])
	while j < CCS_IK_ARRAY_ENTRY_NUM:
		valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+4]
		CCSFilebufferPT += 4
		CCSIKChains[i].append(struct.unpack('<i', valueBuffer)[0])
		j += 1
	i += 1
################################################
################################################



################################################
#           Read CCS vertex info               #
################################################
valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+4]
CCSFilebufferPT += 4
CCS_VERTEX_COUNT = struct.unpack('<i', valueBuffer)[0]


i = 0
while i < CCS_VERTEX_COUNT:
	valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+12]
	CCSFilebufferPT += 12
	CCSVerts.append( ( struct.unpack('<3f', valueBuffer)[0], struct.unpack('<3f', valueBuffer)[1], struct.unpack('<3f', valueBuffer)[2] ) )
	i += 1
################################################
################################################




################################################
#  Read CCS Vertex-to-part index array         #
################################################
i = 0
while i < CCS_VERTEX_COUNT:
	valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+4]
	CCSFilebufferPT += 4
	CCSVertPartArray.append(struct.unpack('<i', valueBuffer)[0])
	i += 1
################################################
################################################


################################################
#      Read CCS Normal array                   #
################################################
i = 0
while i < CCS_VERTEX_COUNT:
	valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+12]
	CCSFilebufferPT += 12
	CCSNormals.append( ( struct.unpack('<3f', valueBuffer)[0], struct.unpack('<3f', valueBuffer)[1], struct.unpack('<3f', valueBuffer)[2] ) )
	i += 1
################################################
################################################



################################################
#      Read CCS eye array                      #
################################################
valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+4]
CCSFilebufferPT += 4
CCS_FACE_COUNT = struct.unpack('<i', valueBuffer)[0]

###Each entry is a dword, so 4 bytes
CCSEyeFacesArrayCopy = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+(CCS_FACE_COUNT*4)]
CCSFilebufferPT += (CCS_FACE_COUNT * 4)

################################################
################################################



################################################
#      Read CCS Face array                     #
################################################

i = 0 
while i < CCS_FACE_COUNT:
	valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+12]
	CCSFilebufferPT += 12
	CCSFaces.append( ( struct.unpack('<3i', valueBuffer)[0], struct.unpack('<3i', valueBuffer)[1], struct.unpack('<3i', valueBuffer)[2] ) )
	i += 1
################################################
################################################



################################################
#        Read CCS bounding boxes               #
################################################
valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+4]
CCSFilebufferPT += 4
CCS_BODY_PART_COUNT = struct.unpack('<i', valueBuffer)[0]


i = 0
while i < CCS_BODY_PART_COUNT:
	valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+24]
	CCSFilebufferPT += 24
	CCSBoundingBox.append( ( struct.unpack('<6f', valueBuffer)[0], struct.unpack('<6f', valueBuffer)[1], struct.unpack('<6f', valueBuffer)[2], struct.unpack('<6f', valueBuffer)[3], struct.unpack('<6f', valueBuffer)[4], struct.unpack('<6f', valueBuffer)[5] ) )
	i += 1
################################################
################################################



################################################
#    Read CCS face group association arrays    #
################################################
facenum = 0

###initialize the 'reverse-lookup' array
i = 0
while i < CCS_FACE_COUNT:
	CCSFaceToPart.append([])
	i += 1


i = 0
while i < CCS_BODY_PART_COUNT:
	CCSFaceGroupArrays.append([])
	valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+4]
	CCSFilebufferPT += 4
	j = struct.unpack('<i', valueBuffer)[0]
	CCSFaceGroupArrays[i].append(j)
	k = 0
	while k < j:
			valueBuffer = CCSFilebuffer[CCSFilebufferPT : CCSFilebufferPT+4]
			CCSFilebufferPT += 4
			facenum = struct.unpack('<i', valueBuffer)[0]
			CCSFaceGroupArrays[i].append( facenum )
			CCSFaceToPart[facenum].append(int(i))
			k += 1
	i += 1
################################################
################################################


orfile.close()

print " done"

################################################
################################################
################################################
################################################






################################################
#   create CSK to CCS relation array           #
################################################
#This simpler import only needs the vertex association
#If the mesh is ever to be edited, will likely also need
#the face relation


print "Calculating CSK to CCS relation . . .",

###initialize arrays
i = 0
while i < CSK_MODEL_VERTEX_COUNT:
	CSKtoCCSVerts.append( int(-1) )
	i += 1

i = 0
while i < CCS_VERTEX_COUNT:
	CCStoCSKVerts_partial.append( int(-1) )
	i += 1



###Create vertex association first. Many vertices will be set more
###than once, but I think it takes just as much processing power to
###check for 'aready set' as it does to just set it again.
###The CCS vertex array will also have values overwritten with
###DIFFERENT values. For the purpose of the array, it only needs one
###valid value, so this is fine, but also makes it not technically
###acurrate
i = 0
while i < CSK_FACE_COUNT:
	CSKtoCCSVerts[CSKConn256[i][5]] = CSKEdgeConn[CSKConn256[i][0]][1]
	CSKtoCCSVerts[CSKConn256[i][6]] = CSKEdgeConn[CSKConn256[i][1]][1]
	CSKtoCCSVerts[CSKConn256[i][7]] = CSKEdgeConn[CSKConn256[i][2]][1]

	CCStoCSKVerts_partial[CSKEdgeConn[CSKConn256[i][0]][1]] = CSKConn256[i][5]
	CCStoCSKVerts_partial[CSKEdgeConn[CSKConn256[i][1]][1]] = CSKConn256[i][6]
	CCStoCSKVerts_partial[CSKEdgeConn[CSKConn256[i][2]][1]] = CSKConn256[i][7]

	i += 1



################################################
################################################


print " done"


################################################
################################################
################################################
################################################





###########################################
#        Read in Base OBJ                 #
###########################################
#Read in the base OBJ file and populate arrays
#These will be needed to calculate the morphs
#as well, since the morphs are displacement values
#based on this base array
#Since it is possible for some OBJ formats to
#include more than one normal or UV coordinate
#per-vertex, this must be accounted for. The
#'collector' arrays below do this by collecting
#all the entries given for a particular vertex.
#The entries are then averaged together (and later
#normalized). REALIZE that this assumes that the
#model has not been 'welded' and is otherwise
#generally 'correct' for the game. If there
#REALLY ARE two UV's or two normals for a vertex,
#this won't work in the game and will be incorrectly
#represented by the simple averaging. The averaging
#is mostly correct and useful in the case where
#an exporter exports the values as 'face' normals
#or calculated slightly different UV for a vertex
#depensing on the face it belongs to (the direction
#from which the program came from when calculating
#the UV)
#The typical case where multiple UV/normal values
#are created is for verticies along UV seams.
#normal or UV for a vertex will be different
#depending on which face is being described
#(the direction from which the vertex is approached)
#This is the reason why more than one vertex will
#be created at a point.

fileLength = 0
facesplit = ''


try:
	orfile = open(files[0][0] + files[0][1], 'rb')
except:
	print
	print "ERROR:Could not open file"
	print files[0][0]
	print "for reading."
	print "Why? File seems to exist. Persmission problems?"
	print
	owfileCSK.close()
	owfileCCS.close()
	raise SystemExit

print "Reading base model OBJ . . .",


OBJFileBuffer = orfile.readlines()


###Initialize array with correct number of 'buckets'
i = 0
while i < CSK_MODEL_VERTEX_COUNT:
	OBJUVcollector.append( [] )
	OBJNormalcollector.append( [] )
	i += 1

fileLength = len(OBJFileBuffer)
i = 0
while i < fileLength:
	valueBuffer = OBJFileBuffer[i]
	valueBuffer = valueBuffer.strip()
	if valueBuffer == "":
		pass
	else:
		###Remember, when looking for the length of the split string, the leading letter
		###is still part of the string, if there should be three values, there will be
		###four entries after the split
		valueBuffer = valueBuffer.split()
		if valueBuffer[0] == " " or valueBuffer[0] == "\n":
			pass
		elif valueBuffer[0] == "v":
			if len(valueBuffer) != 4:
				print
				print "ERROR:OBJ formatting error in"
				print files[0][0]+files[0][1]
				print "vertex number "+str(len(OBJVertex)+1)+" bad"
				print "Too many or too few values (3)?"
				print "Odd whitespace?"
				print
				owfileCSK.close()
				owfileCCS.close()
				orfile.close()
				raise SystemExit
			OBJVertex.append( ( float(valueBuffer[1]), float(valueBuffer[2]), float(valueBuffer[3])*NVZ ) )
		elif valueBuffer[0] == "vn":
			if len(valueBuffer) != 4:
				print
				print "ERROR:OBJ formatting error in"
				print files[0][0]+files[0][1]
				print "normal number "+str(len(OBJNormalTemp)+1)+" bad"
				print "Too many or too few values (3)?"
				print "Odd whitespace?"
				print
				owfileCSK.close()
				owfileCCS.close()
				orfile.close()
				raise SystemExit
			OBJNormalTemp.append( ( float(valueBuffer[1]), float(valueBuffer[2]), float(valueBuffer[3])*NNZ ) )
		elif valueBuffer[0] == "vt":
			if len(valueBuffer) != 3:
				print
				print "ERROR:OBJ formatting error in"
				print files[0][0]+files[0][1]
				print "UV number "+str(len(OBJUVTemp)+1)+" bad"
				print "Too many or too few values (2)?"
				print "Odd whitespace?"
				print
				owfileCSK.close()
				owfileCCS.close()
				orfile.close()
				raise SystemExit
			OBJUVTemp.append( ( float(valueBuffer[1]), float(valueBuffer[2]) ) )
		elif valueBuffer[0] == "f":
			if len(valueBuffer) != 4:
				print
				print "ERROR:OBJ formatting error in"
				print files[0][0]+files[0][1]
				print "face number "+str(len(OBJFaces)+1)+" bad"
				print "Too many or too few vertices (3)?"
				print "Odd whitespace?"
				print "B&W2 may only use trianlges"
				print
				owfileCSK.close()
				owfileCCS.close()
				orfile.close()
				raise SystemExit
			facesplit = valueBuffer[1].split("/") + valueBuffer[2].split("/") + valueBuffer[3].split("/")
			if len(facesplit) != 9:
				print
				print "ERROR:OBJ formatting error in"
				print files[0][0]+files[0][1]
				print "face number "+str(len(OBJFaces)+1)+" bad"
				print "This importer requires both normals and"
				print "UV points for each face vertex"
				print " ( format: #/#/# )"
				print
				owfileCSK.close()
				owfileCCS.close()
				orfile.close()
				raise SystemExit
			###Currently, the OBJFaces array is unused in this script. It is NOT caluculated
			###correctly for creating the face array, either. Remember, every even-numbered
			###entry should be re-arranged to be placed into the face array in the 'correct'
			###order for a trinagle strip
			OBJFaces.append( ( facesplit[0] , facesplit[3] , facesplit[6] ) )

			OBJUVcollector[int(facesplit[0])-1].append( OBJUVTemp[int(facesplit[1])-1] )
			OBJUVcollector[int(facesplit[3])-1].append( OBJUVTemp[int(facesplit[4])-1] )
			OBJUVcollector[int(facesplit[6])-1].append( OBJUVTemp[int(facesplit[7])-1] )
			OBJNormalcollector[int(facesplit[0])-1].append( OBJNormalTemp[int(facesplit[2])-1] )
			OBJNormalcollector[int(facesplit[3])-1].append( OBJNormalTemp[int(facesplit[5])-1] )
			OBJNormalcollector[int(facesplit[6])-1].append( OBJNormalTemp[int(facesplit[8])-1] )
		else:
			###There are a number of other things that can be in an OBJ, like nurbs, etc..
			###just ignore them I guess.
			pass
	i += 1

################################################
################################################

print " done"

orfile.close()


################################################
################################################





###########################################
#        Calculate new model data         #
###########################################

print "Calculating new model data .",

###########################################
#        Average Normals and UVs          #
###########################################
#Some OBJs will have multiple normals/UVs per vertex, OR will
#describe normals per-face instead of per-vertex (blender). I
#believe normals in OBJ should always be per-vertex, but..
#Average the values. See notes above on previous section.
Vx = 0.0
Vy = 0.0
Vz = 0.0

i = 0
while i < CSK_MODEL_VERTEX_COUNT:
	Vx = 0.0
	Vy = 0.0
	j = 0
	k = len(OBJUVcollector[i])
	while j < k:
		Vx += OBJUVcollector[i][j][0]
		Vy += OBJUVcollector[i][j][1]
		j += 1
	###If the UV coordinates are to be un-arranged
	if NUV:
		if CSKUV[i][0] > 0.5:
			OBJUV.append( ( (1-((Vx/len(OBJUVcollector[i]))/2)) , (1-(Vy/len(OBJUVcollector[i]))) ) )
		else:
			OBJUV.append( ( (Vx/len(OBJUVcollector[i]))/2 , (1-(Vy/len(OBJUVcollector[i]))) ) )
	else:
		OBJUV.append( ( Vx / len(OBJUVcollector[i]) , Vy / len(OBJUVcollector[i]) ) )
	Vx = 0.0
	Vy = 0.0
	Vz = 0.0
	j = 0
	k = len(OBJNormalcollector[i])
	while j < k:
		Vx += OBJNormalcollector[i][j][0]
		Vy += OBJNormalcollector[i][j][1]
		Vz += OBJNormalcollector[i][j][2]
		j += 1
	###The resulting normal will be normalized below
	OBJNormals.append( ( Vx / len(OBJNormalcollector[i]) , Vy / len(OBJNormalcollector[i]) , Vz / len(OBJNormalcollector[i]) ) )
	i += 1

print ".",

###########################################
#        Create duplicate vertex array    #
###########################################
###tempArray will keep a temporary list of duplicate verts


i = 0
while i < CSK_MODEL_VERTEX_COUNT:
	OBJVertexDuplicates.append( [] )
	i += 1

i = 0
while i < CSK_MODEL_VERTEX_COUNT:
	###If this vertex has not already had its duplicates found
	if len(OBJVertexDuplicates[i]) == 0:
		tempArray = []
		###Only have to check vertices not already passed over by main loop
		j = i + 1
		while j < CSK_MODEL_VERTEX_COUNT:
			if (abs(OBJVertex[i][0]-OBJVertex[j][0]) < 0.0002) and (abs(OBJVertex[i][1]-OBJVertex[j][1]) < 0.0002) and (abs(OBJVertex[i][2]-OBJVertex[j][2]) < 0.0002):
				tempArray.append ( int(j) )
			j += 1
		###If duplicates were found
		if len(tempArray) > 0:
			###The vertex index represented by i can just be assigned the
			###array of duplicates. The others have to be handled
			OBJVertexDuplicates[i] = tempArray
			j = 0
			###K will be the number of duplicate verts found for this vert (i)
			k = len(tempArray)
			while j < k:
				###Append the firt index, then append the others as long as they
				###aren't this verticies own index
				OBJVertexDuplicates[tempArray[j]].append( int(i) )
				m = 0
				while m < k:
					if m != j:
						OBJVertexDuplicates[tempArray[j]].append( int(tempArray[m]) )
					m += 1
				j += 1
	i += 1

print ".",

###########################################
#        Normalize Normals                #
###########################################
#Normalize = make a unit vector
#Average and Normalize normal vectors for duplicate vertices
#This still produces values different from the original models
#contain. They are pretty close and good enough, I think..
#Upon further testing, it is basically perfect, except for
#vertices around the eyes where normals were purposefully made
#slightly non-normal.. don't know why currently.
Vx = 0.0
Vy = 0.0
Vz = 0.0
length = 0.0
#Will be the number of duplicate vertices a vertex has
numDuplicates = 0

i = 0 
while i < CSK_MODEL_VERTEX_COUNT:
	OBJNormalizedNormals.append([])
	OBJNormalizedNormals[i].append(0.0)
	OBJNormalizedNormals[i].append(0.0)
	OBJNormalizedNormals[i].append(0.0)
	i += 1
i = 0 
while i < CSK_MODEL_VERTEX_COUNT:
	###If the normal has not yet been set
	if (OBJNormalizedNormals[i][0] == 0.0) and (OBJNormalizedNormals[i][1] == 0.0) and (OBJNormalizedNormals[i][2] == 0.0):
		numDuplicates = len(OBJVertexDuplicates[i])
		length = 0.0
		###Add current vertex data
		Vx = OBJNormals[i][0]
		Vy = OBJNormals[i][1]
		Vz = OBJNormals[i][2]
		###Add any duplicate vertex data
		j = 0
		while j < numDuplicates:
			Vx += OBJNormals[OBJVertexDuplicates[i][j]][0]
			Vy += OBJNormals[OBJVertexDuplicates[i][j]][1]
			Vz += OBJNormals[OBJVertexDuplicates[i][j]][2]
			j += 1
		###(+1) because the current vertex counts as 1, plus the number of duplicates
		Vx = Vx / (numDuplicates+1)
		Vy = Vy / (numDuplicates+1)
		Vz = Vz / (numDuplicates+1)
		length = math.sqrt( (Vx*Vx) + (Vy*Vy) + (Vz*Vz) )
		Vx = Vx / length
		Vy = Vy / length
		Vz = Vz / length

		###Set current vertex's data
		OBJNormalizedNormals[i][0] = Vx
		OBJNormalizedNormals[i][1] = Vy
		OBJNormalizedNormals[i][2] = Vz
		###Set any duplicate vertex's data
		j = 0
		while j < numDuplicates:
			OBJNormalizedNormals[OBJVertexDuplicates[i][j]][0] = Vx
			OBJNormalizedNormals[OBJVertexDuplicates[i][j]][1] = Vy
			OBJNormalizedNormals[OBJVertexDuplicates[i][j]][2] = Vz
			j += 1
	i += 1


print ".",

###########################################
#   Calculate Tangent and Bitangent       #
###########################################
#For now, the second set of tan/bitan are simply
#set to the first set. This is correct everywhere
#except for verts along the UV seams. These seem
#to have different values among the two sets.
tv = []
bv = []
V1 = 0
V2 = 0
V3 = 0
x1 = 0
x2 = 0
y1 = 0
y2 = 0
z1 = 0
z2 = 0
s1 = 0
s2 = 0
t1 = 0
t2 = 0
r = 0
dotNT = 0
dotNB = 0
TX = 0
TY = 0
TZ = 0
BX = 0
BY = 0
BZ = 0
TLength = 0
BLength = 0


i = 0
while i < CSK_MODEL_VERTEX_COUNT:
	tv = [ 0.0 , 0.0 , 0.0 ]
	bv = [ 0.0 , 0.0 , 0.0 ]
	k = len(CSKVertexFaceIndex[i])
	j = 0
	###For each face that contains this vertex...
	while j < k:
		###V1 is the vertex INDEX of this vertex
		###V2 and V3 are the indexes of the other vertices in this face
		V1 = i
		if CSKFaces[CSKVertexFaceIndex[i][j]][0] == i:
			V2 = CSKFaces[CSKVertexFaceIndex[i][j]][1]
			V3 = CSKFaces[CSKVertexFaceIndex[i][j]][2]
		elif CSKFaces[CSKVertexFaceIndex[i][j]][1] == i:
			V2 = CSKFaces[CSKVertexFaceIndex[i][j]][0]
			V3 = CSKFaces[CSKVertexFaceIndex[i][j]][2]
		else:
			V2 = CSKFaces[CSKVertexFaceIndex[i][j]][0]
			V3 = CSKFaces[CSKVertexFaceIndex[i][j]][1]
		x1 = OBJVertex[V2][0] - OBJVertex[V1][0]
		x2 = OBJVertex[V3][0] - OBJVertex[V1][0]
		y1 = OBJVertex[V2][1] - OBJVertex[V1][1]
		y2 = OBJVertex[V3][1] - OBJVertex[V1][1]
		z1 = OBJVertex[V2][2] - OBJVertex[V1][2]
		z2 = OBJVertex[V3][2] - OBJVertex[V1][2]
		if CSKUV[i][0] > 0.5:
			s1 = ((1-(OBJUV[V2][0]))*2) - ((1-(OBJUV[V1][0]))*2)
			s2 = ((1-(OBJUV[V3][0]))*2) - ((1-(OBJUV[V1][0]))*2)
			t1 = (-OBJUV[V2][1]) + OBJUV[V1][1]
			t2 = (-OBJUV[V3][1]) + OBJUV[V1][1]
		else:
			s1 = (OBJUV[V2][0]*2) - (OBJUV[V1][0]*2)
			s2 = (OBJUV[V3][0]*2) - (OBJUV[V1][0]*2)
			t1 = (-OBJUV[V2][1]) + OBJUV[V1][1]
			t2 = (-OBJUV[V3][1]) + OBJUV[V1][1]
		r = 1.0 / ((s1*t2-s2*t1) + 0.00001)
		tv[0] += (t2*x1-t1*x2)*r
		tv[1] += (t2*y1-t1*y2)*r
		tv[2] += (t2*z1-t1*z2)*r
		bv[0] += (s1*x2-s2*x1)*r
		bv[1] += (s1*y2-s2*y1)*r
		bv[2] += (s1*z2-s2*z1)*r
		j += 1
	dotNT = (OBJNormalizedNormals[i][0]*tv[0]) + (OBJNormalizedNormals[i][1]*tv[1]) + (OBJNormalizedNormals[i][2]*tv[2])
	TX = tv[0] - (OBJNormalizedNormals[i][0] * dotNT)
	TY = tv[1] - (OBJNormalizedNormals[i][1] * dotNT)
	TZ = tv[2] - (OBJNormalizedNormals[i][2] * dotNT)

	###Misnomer here.. not length yet. The calculation really does call for the length, but it will
	###be squared anyway, no need to sqrt the length here yet
	TLength = (TX*TX) + (TY*TY) + (TZ*TZ)

	dotNB = (OBJNormalizedNormals[i][0]*bv[0]) + (OBJNormalizedNormals[i][1]*bv[1]) + (OBJNormalizedNormals[i][2]*bv[2])
	dotTpB = (TX*bv[0]) + (TY*bv[1]) + (TZ*bv[2])
	###Handedness alway the opposite of this calculation
	BX = -(bv[0] - (OBJNormalizedNormals[i][0] * dotNB) - (( TX * dotTpB) / TLength))
	BY = -(bv[1] - (OBJNormalizedNormals[i][1] * dotNB) - (( TY * dotTpB) / TLength))
	BZ = -(bv[2] - (OBJNormalizedNormals[i][2] * dotNB) - (( TZ * dotTpB) / TLength))


	BLength = math.sqrt( (BX*BX) + (BY*BY) + (BZ*BZ) )
	###Now make it the real length value
	TLength = math.sqrt(TLength)

	###Some lengths actaully calculate out to be 0. I'm not sure why
	###But nothing seems to break. Just make sure no divide by 0.
	if BLength == 0:
		BX = 0.0
		BY = 0.0
		BZ = 0.0
	else:
		BX = BX / BLength
		BY = BY / BLength
		BZ = BZ / BLength

	if TLength == 0:
		TX = 0.0
		TY = 0.0
		TZ = 0.0
	else:		
		TX = TX / TLength
		TY = TY / TLength
		TZ = TZ / TLength


	OBJTangents.append( ( TX , TY , TZ ) )
	OBJBitangents.append( ( BX , BY , BZ ) )
	i += 1

print " done"

################################################
################################################






################################################
#        Begin Creating new CSK file           #
################################################

print "Creating new base CSK model . . .",

################################################
#              Create CSK header               #
################################################
#Only create the radeon section, skip the GeForce section
#The geforce section must still be 'correct', however
scratchvalue = "LiOnHeAdCreature"
scratchvalue += "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
scratchvalue += struct.pack('<i', 12)
scratchvalue += struct.pack('<i', 4663)
scratchvalue += struct.pack('<i', 3)

###No ability to create new bones at this time. Copy all bone info
scratchvalue += struct.pack('<i', CSK_BODY_PART_MAX)

###Now, I'm assuming that the GeForce section can be ignored.
scratchvalue += "geforce"
scratchvalue += "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
scratchvalue += struct.pack('<i', 60)
scratchvalue += struct.pack('<i', 0)
scratchvalue += struct.pack('<i', 0)
scratchvalue += struct.pack('<i', 0)
scratchvalue += struct.pack('<i', 0)
scratchvalue += "base"
scratchvalue += "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
scratchvalue += struct.pack('<i', 0)
scratchvalue += struct.pack('<i', 0)
scratchvalue += struct.pack('<i', 0)

###Now, begin the radeon section
scratchvalue += "radeon"
scratchvalue += "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"

###This is the section size. Since it is UNKOWN, it will need to be written later.
###To facilitate this, the buffer accumulation is written out now, and the write-
###file position is saved. It will be returned to later and written
owfileCSK.seek(0)
owfileCSK.write(scratchvalue)
scratchvalue = ''
NewCSKwriteFileSectionSizeLoc = owfileCSK.tell()
scratchvalue += struct.pack('<i', 1234)



################################################
#        Create new CSK material section       #
################################################
#Since so many of the values here are related to
#bones, most of the info (all of it for now) is
#simply copied from the original CSK.
NUM_PART_ARRAYS = 0
PART_ARRAY_COUNT = 0

###The material section has an apparent 80 body-part
###limit.. a good deal of extra work must be done to
###get around this..
scratchvalue += struct.pack('<i', len(CSKMats)/4)

i = 0
while i < len(CSKMats)/4:
	scratchvalue += struct.pack('<i', 1)
	scratchvalue += CSKMats[(i*4)][1]
	scratchvalue += CSK384byteArray
	NUM_PART_ARRAYS = CSKMats[(i*4)][0]
	scratchvalue += struct.pack('<i', NUM_PART_ARRAYS)
	j = 0
	while j < NUM_PART_ARRAYS:
		PART_ARRAY_COUNT = CSKMats[(i*4)+2][(j*3)]
		scratchvalue += struct.pack('<i', PART_ARRAY_COUNT)
		k = 0
		while k < PART_ARRAY_COUNT:
			scratchvalue += struct.pack('<i', CSKMats[(i*4)+1][j][k])
			k += 1
		scratchvalue += struct.pack('<i', CSKMats[(i*4)+2][(j*3)+1])
		scratchvalue += struct.pack('<i', CSKMats[(i*4)+2][(j*3)+2])
		j += 1
	i += 1

	
################################################
#        Skip Face Array size definition       #
################################################
#Since the length of the face array will likely change, this value must be writen out after creating
#the face array. Get the output file position of this number to return later
# NOTE: This is sort of untrue since the face array cannot be edited now.. it could be writen out
#       now since the array is simply being copied. It will just be done further in the scripts as
#       indicated, but will just be the length of the face array / 2 (each entry is 2 bytes).

owfileCSK.write(scratchvalue)
scratchvalue = ''
NewCSKfaceArrayCountLoc = owfileCSK.tell()
scratchvalue += struct.pack('<i', 1234)


################################################
#          Create new Face Array               #
################################################
#Generate the new face array. As far as I can tell, this is a
#"Degenreative Triangle Strip" array that defines the entire
#model, except split by the body part-part defenitions in
#the material section.
# NOTE: For now, just copy the face array. No ability to change
#       bones, so no ability to add/remove vertices, so there
#       can be no change to the face structure. Since there can
#       be no change, no reason to recalculate the whole strip
faceArraycounter = 0

scratchvalue += CSKFaceArrayCopy


################################################
#      Write Face Array size definition        #
################################################
#This is somewhat 'incorrect' for now since the face array
#is just copied

owfileCSK.write(scratchvalue)
scratchvalue = ''

###faceArraycounter is never set since the face array is just
###copied for now
#scratchvalue += struct.pack('<i', faceArraycounter)
### "/2" because each entry is two bytes, and this number wants the
###entry count, not size.
scratchvalue += struct.pack('<i', len(CSKFaceArrayCopy)/2 )

tempLoc = owfileCSK.tell()
owfileCSK.seek(NewCSKfaceArrayCountLoc)
owfileCSK.write(scratchvalue)
scratchvalue = ''
owfileCSK.seek(tempLoc)
################################################
################################################



###########################################
#      write out base UV/metadata array   #
###########################################
#Note that this exports one normal/UV per vertex


###UV/metadata array
scratchvalue += struct.pack('<i', CSK_VERT_META_ARRAY_SUBARRAY_SIZE)
###Since this importer doesn't allow changing the vertex number, array size
###will remain the same as in the CSK file. No need to recalculate
scratchvalue += struct.pack('<i', CSK_VERT_META_ARRAY_SUBARRAY_SIZE * CSK_MODEL_VERTEX_COUNT)

i = 0
while i < CSK_MODEL_VERTEX_COUNT:
	scratchvalue += struct.pack('<f', OBJUV[i][0]) + struct.pack('<f', OBJUV[i][1])
	###remember, one word of padding, thus the '0'.
	###The bone weights are simply copied for now.. the below is wrong, anyway. To really get
	###the correct values for the bones, the entire structure of the body part arrays would have
	###to be parsed again.
	scratchvalue += struct.pack('<H', CSKBoneWeights[i][0]) + struct.pack('<H', CSKBoneWeights[i][2]) + struct.pack('<H', CSKBoneWeights[i][4]) + struct.pack('<H', 0) + struct.pack('<f', CSKBoneWeights[i][1]) + struct.pack('<f', CSKBoneWeights[i][3]) + struct.pack('<f', CSKBoneWeights[i][5])
	i += 1


owfileCSK.write(scratchvalue)
scratchvalue = ''


###########################################
#      write out base Vertex array        #
###########################################
scratchvalue += "base"
scratchvalue += "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
###Vertex sub-array size should be 72 bytes (18 floats)
scratchvalue += struct.pack('<i', 72)
###The size of this array will be CSK_MODEL_VERTEX_COUNT*72
scratchvalue += struct.pack('<i', CSK_MODEL_VERTEX_COUNT * 72)

i = 0
while i < CSK_MODEL_VERTEX_COUNT:
	scratchvalue += struct.pack('<f', OBJVertex[i][0]) + struct.pack('<f', OBJVertex[i][1]) + struct.pack('<f', OBJVertex[i][2]) + struct.pack('<f', OBJNormalizedNormals[i][0]) + struct.pack('<f', OBJNormalizedNormals[i][1]) + struct.pack('<f', OBJNormalizedNormals[i][2]) + struct.pack('<f', OBJTangents[i][0]) + struct.pack('<f', OBJTangents[i][1]) + struct.pack('<f', OBJTangents[i][2]) + struct.pack('<f', OBJBitangents[i][0]) + struct.pack('<f', OBJBitangents[i][1]) + struct.pack('<f', OBJBitangents[i][2]) + struct.pack('<f', OBJTangents[i][0]) + struct.pack('<f', OBJTangents[i][1]) + struct.pack('<f', OBJTangents[i][2]) + struct.pack('<f', OBJBitangents[i][0]) + struct.pack('<f', OBJBitangents[i][1]) + struct.pack('<f', OBJBitangents[i][2])
	i += 1

owfileCSK.write(scratchvalue)
scratchvalue = ''

################################################
################################################

print " done"


################################################
################################################









################################################
#        Begin Creating new CCS file           #
################################################

print "Creating new CCS model. . .",

################################################
#              Create CCS header               #
################################################
scratchvalue = "LiOnHeAdCreature"
scratchvalue += "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"


###This is the file size. Since it is UNKOWN, it will need to be written later.
###To facilitate this, the buffer accumulation is written out now, and the write-
###file position is saved. It will be returned to later and written
owfileCCS.seek(0)
owfileCCS.write(scratchvalue)
scratchvalue = ''
NewCCSwriteFileSizeLoc = owfileCCS.tell()
scratchvalue += struct.pack('<i', int(0) )


################################################
#              Create CCS IK chains            #
################################################
#These are just copied for now

###This is the number of entries in the IK chains (the number of chains
### * the length of the chains) plus the two length count numbers
scratchvalue += struct.pack( '<i', ((len(CCSIKChains)*len(CCSIKChains[0]))+2) )

k = len(CCSIKChains)
scratchvalue += struct.pack('<i', k)
###All the IK chains should have the same length, so just pick one
m = len(CCSIKChains[0])
scratchvalue += struct.pack('<i', m)
i = 0
while i < k:
	j = 0
	while j < m:
		scratchvalue += struct.pack('<i', CCSIKChains[i][j])
		j += 1
	i += 1

################################################
#              Create CCS vertex array         #
################################################
#Since the mesh structure cannot be edited, should
#have the same number of vertices as the original CCS
#This uses the Edge sections in the CSK file to find
#the relation of the CSK vertices to the CCS vertices
#and sets the CCS vertices here. The importing
#model in the OBJ SHOULD be a 'CSK' model


###Write CCS vertex count
scratchvalue += struct.pack('<i', CCS_VERTEX_COUNT)



###Use the CCStoCSKVerts_partial array to associate the CCS vertex with a
###CSK vertex and write out it's X Y Z values, but read from the importing
###model (OBJ) not the original model (CSK)...
i = 0
while i < CCS_VERTEX_COUNT:
	scratchvalue += struct.pack('<f', OBJVertex[CCStoCSKVerts_partial[i]][0]) + struct.pack('<f', OBJVertex[CCStoCSKVerts_partial[i]][1]) + struct.pack('<f', OBJVertex[CCStoCSKVerts_partial[i]][2])
	i += 1

################################################
#   Create CCS vertex part association array   #
################################################

i = 0
while i < CCS_VERTEX_COUNT:
	scratchvalue += struct.pack('<i', CCSVertPartArray[i])
	i += 1

################################################
#   Create CCS vertex normal array             #
################################################

###Like the vertex array above, use the CCS to CSK association
i = 0
while i < CCS_VERTEX_COUNT:
	scratchvalue += struct.pack('<f', OBJNormalizedNormals[CCStoCSKVerts_partial[i]][0]) + struct.pack('<f', OBJNormalizedNormals[CCStoCSKVerts_partial[i]][1]) + struct.pack('<f', OBJNormalizedNormals[CCStoCSKVerts_partial[i]][2])
	i += 1


################################################
#      Create the eye array                    #
################################################
#Since the mesh cannot be modified, simply enter the copied array

###The CCS model face count
scratchvalue += struct.pack('<i', CCS_FACE_COUNT)

scratchvalue += CCSEyeFacesArrayCopy



################################################
#      Create the CCS Face Array               #
################################################
#Unlike the CSK model, this model has the faces stored
#in order and in an easy to manage format. For now, also
#copy the faces array

i = 0
while i < CCS_FACE_COUNT:
	scratchvalue += struct.pack('<i', CCSFaces[i][0]) + struct.pack('<i', CCSFaces[i][1]) + struct.pack('<i', CCSFaces[i][2])
	i += 1



################################################
#      Create the CCS Bounding Box Array       #
################################################
#The bounding box array calculation looks like it's going
#to require bone information. I suspect that, to correctly
#calculate the boxes in the way the game expects, each
#part will need to be transformed into its 'default'
#position by a transformation matrix, probably found
#in the CBN file.
#For now, just copy it.

###The CCS model body part count
scratchvalue += struct.pack('<i', CCS_BODY_PART_COUNT)

###Copy the array
i = 0
while i < CCS_BODY_PART_COUNT:
	scratchvalue += struct.pack('<6f', CCSBoundingBox[i][0], CCSBoundingBox[i][1], CCSBoundingBox[i][2], CCSBoundingBox[i][3], CCSBoundingBox[i][4], CCSBoundingBox[i][5])
	i += 1


################################################
#Create the CCS face-to-part association arrays#
################################################
#Once again, Since the mesh cannot be edited, simply
#copy the face-part association arrays
partCount = 0

i = 0
while i < CCS_BODY_PART_COUNT:
	partCount = CCSFaceGroupArrays[i][0]
	scratchvalue += struct.pack('<i', partCount)
	j = 1
	while j <= partCount:
		scratchvalue += struct.pack('<i', CCSFaceGroupArrays[i][j])
		j +=1
	i +=1



owfileCCS.write(scratchvalue)
scratchvalue = ''



################################################
#      Write the new CCS file filesize         #
################################################

tempLoc = owfileCCS.tell()
owfileCCS.seek(NewCCSwriteFileSizeLoc)
### (-4) since the size doesn't include this number itself
owfileCCS.write( struct.pack('<i', (tempLoc-NewCCSwriteFileSizeLoc)-4) )


################################################
################################################

owfileCCS.close()

print " done"

################################################
################################################










###########################################
#    Retrieve and generate morph data     #
###########################################
#A number of arrays are used, but are declared below
#since they need to be reset inside the morph loop.
#These are bsically the same structure as used to
#read in the base model above
#
#Long section.. read in the OBJ information, gather the vertices, normal, and UV info in the same
#way as done above for the base model. This is just a copy-paste of the code.
#After reading in, determine which vertices have moved. Create a new vertex array of only the
#vertices that have been displaced and alter the vertex information to be the displacement
#value (positive or negative in all 3D). The Normal and Tan/Bin values will also need to be
#recalculated. These values are also specified as displacements. The UV info from the morph
#OBJ is needed for calculations, but is not actually re-written anywhere in the CSK.
morphCount = 0

print "Creating new CSK morphs : ",

###Quick CSK data write.. the number of morphs
scratchvalue += struct.pack('<i', CSK_MORPH_COUNT)
owfileCSK.write(scratchvalue)
scratchvalue = ''

###The number of morphs specified on the command line
morphCount = len(files) - 1

u = 0
while u < morphCount:
	####VARS
	OBJMorphUVcollector = []
	OBJMorphNormalcollector = []
	OBJMorphVertex = []
	OBJMorphNormalTemp = []
	OBJMorphUVTemp = []
	OBJMorphFaces = []
	OBJMorphUV = []
	OBJMorphNormals = []
	OBJMorphNormalizedNormals = []
	OBJMorphTangents = []
	OBJMorphBitangents = []
	OBJMorphVertexDisplacement = []
	OBJMorphBitangentsDisplacement = []
	OBJMorphTangentsDisplacement = []
	OBJMorphNormalizedNormalsDisplacement = []


	###u + 1 since index 0 is the base morph.. dirty, well crap nuggets
	try:
		orfile = open(files[u+1][0] + files[u+1][1], 'rb')
	except:
		print
		print "ERROR:Could not open OBJ morph file"
		print str(files[u+1][0]) + str(files[u+1][1])
		print "Why? File seems to exist. Persmission problems?"
		print
		owfileCSK.close()
		raise SystemExit

	print str(files[u+1][2]),


###########################################
#       Read in morph OBJ file            #
###########################################
	OBJFileBuffer = orfile.readlines()
	i = 0
	while i < CSK_MODEL_VERTEX_COUNT:
		OBJMorphUVcollector.append( [] )
		OBJMorphNormalcollector.append( [] )
		i += 1
	i = 0
	while i < len(OBJFileBuffer):
		valueBuffer = OBJFileBuffer[i]
		valueBuffer = valueBuffer.strip()
		if valueBuffer == "":
			pass
		else:
			###Remember, when looking for the length of the split string, the leading letter
			###is still part of the string, if there should be three values, there will be
			###four entries after the split
			valueBuffer = valueBuffer.split()
			if valueBuffer[0] == " " or valueBuffer[0] == "\n":
				pass
			elif valueBuffer[0] == "v":
				if len(valueBuffer) != 4:
					print
					print "ERROR:OBJ formatting error in"
					print files[u+1][0]+files[u+1][1]
					print "vertex number "+str(len(OBJMorphVertex)+1)+" bad"
					print "Too many or too few values (3)?"
					print "Odd whitespace?"
					print
					owfileCSK.close()
					orfile.close()
					raise SystemExit
				OBJMorphVertex.append( ( float(valueBuffer[1]), float(valueBuffer[2]), float(valueBuffer[3])*NVZ ) )
			elif valueBuffer[0] == "vn":
				if len(valueBuffer) != 4:
					print
					print "ERROR:OBJ formatting error in"
					print files[u+1][0]+files[u+1][1]
					print "normal number "+str(len(OBJMorphNormalTemp)+1)+" bad"
					print "Too many or too few values (3)?"
					print "Odd whitespace?"
					print
					owfileCSK.close()
					orfile.close()
					raise SystemExit
				OBJMorphNormalTemp.append( ( float(valueBuffer[1]), float(valueBuffer[2]), float(valueBuffer[3])*NNZ ) )
			elif valueBuffer[0] == "vt":
				if len(valueBuffer) != 3:
					print
					print "ERROR:OBJ formatting error in"
					print files[u+1][0]+files[u+1][1]
					print "UV number "+str(len(OBJMorphUVTemp)+1)+" bad"
					print "Too many or too few values (2)?"
					print "Odd whitespace?"
					print
					owfileCSK.close()
					orfile.close()
					raise SystemExit
				OBJMorphUVTemp.append( ( float(valueBuffer[1]), float(valueBuffer[2]) ) )
			elif valueBuffer[0] == "f":
				if len(valueBuffer) != 4:
					print
					print "ERROR:OBJ formatting error in"
					print files[u+1][0]+files[u+1][1]
					print "face number "+str(len(OBJMorphFaces)+1)+" bad"
					print "Too many or too few vertices (3)?"
					print "Odd whitespace?"
					print "B&W2 may only use trianlges"
					print
					owfileCSK.close()
					orfile.close()
					raise SystemExit
				facesplit = valueBuffer[1].split("/") + valueBuffer[2].split("/") + valueBuffer[3].split("/")
				if len(facesplit) != 9:
					print
					print "ERROR:OBJ formatting error in"
					print files[u+1][0]+files[u+1][1]
					print "face number "+str(len(OBJMorphFaces)+1)+" bad"
					print "This importer requires both normals and"
					print "UV points for each face vertex"
					print " ( format: #/#/# )"
					print
					owfileCSK.close()
					orfile.close()
					raise SystemExit
				###Currently, the OBJFaces array is unused in this script. It is NOT caluculated
				###correctly for creating the face array, either. Remember, every even-numbered
				###entry should be re-arranged to be placed into the face array in the 'correct'
				###order for a trinagle strip
				OBJMorphFaces.append( ( facesplit[0] , facesplit[3] , facesplit[6] ) )

				OBJMorphUVcollector[int(facesplit[0])-1].append( OBJMorphUVTemp[int(facesplit[1])-1] )
				OBJMorphUVcollector[int(facesplit[3])-1].append( OBJMorphUVTemp[int(facesplit[4])-1] )
				OBJMorphUVcollector[int(facesplit[6])-1].append( OBJMorphUVTemp[int(facesplit[7])-1] )
				OBJMorphNormalcollector[int(facesplit[0])-1].append( OBJMorphNormalTemp[int(facesplit[2])-1] )
				OBJMorphNormalcollector[int(facesplit[3])-1].append( OBJMorphNormalTemp[int(facesplit[5])-1] )
				OBJMorphNormalcollector[int(facesplit[6])-1].append( OBJMorphNormalTemp[int(facesplit[8])-1] )
			else:
				pass
		i += 1

	###Added for morphs specifically: If the number of vertices in this model does not match the base model
	###Something is wrong
	if len(OBJMorphVertex) != CSK_MODEL_VERTEX_COUNT:
		print
		print "ERROR:Model error in"
		print files[u+1][0]+files[u+1][1]
		print "This morph model has a different number of"
		print "vertices than the base model. It should have"
		print str(CSK_MODEL_VERTEX_COUNT)
		print "vertices, but instead has"
		print str(len(OBJMorphVertex))
		print
		owfileCSK.close()
		orfile.close()
		raise SystemExit

###########################################
#        Average morph Normals and UVs    #
###########################################
	Vx = 0.0
	Vy = 0.0
	Vz = 0.0

	i = 0
	while i < CSK_MODEL_VERTEX_COUNT:
		Vx = 0.0
		Vy = 0.0
		j = 0
		k = len(OBJMorphUVcollector[i])
		while j < k:
			Vx += OBJMorphUVcollector[i][j][0]
			Vy += OBJMorphUVcollector[i][j][1]
			j += 1
		###If the UV coordinates are to be un-arranged
		if NUV:
			if CSKUV[i][0] > 0.5:
				OBJMorphUV.append( ( (1-((Vx/len(OBJMorphUVcollector[i]))/2)) , (1-(Vy/len(OBJMorphUVcollector[i]))) ) )
			else:
				OBJMorphUV.append( ( (Vx/len(OBJMorphUVcollector[i]))/2 , (1-(Vy/len(OBJMorphUVcollector[i]))) ) )
		else:
			OBJMorphUV.append( ( Vx / len(OBJMorphUVcollector[i]) , Vy / len(OBJMorphUVcollector[i]) ) )
		Vx = 0.0
		Vy = 0.0
		Vz = 0.0
		j = 0
		k = len(OBJMorphNormalcollector[i])
		while j < k:
			Vx += OBJMorphNormalcollector[i][j][0]
			Vy += OBJMorphNormalcollector[i][j][1]
			Vz += OBJMorphNormalcollector[i][j][2]
			j += 1
		OBJMorphNormals.append( ( Vx / len(OBJMorphNormalcollector[i]) , Vy / len(OBJMorphNormalcollector[i]) , Vz / len(OBJMorphNormalcollector[i]) ) )
		i += 1

###########################################
#        Normalize Normals                #
###########################################
#The determiniation of which vertecies are duplicates
#was found when reading in the base model array
#The same duplicate verticies should still be
#duplicates in the morphs. If two vertices occupy similar
#space NOW that didn't in the base, it very likely is
#because the morph brought them together rather than
#them begin duplicates for a technical reason (UV seams
#or other)

	Vx = 0.0
	Vy = 0.0
	Vz = 0.0
	length = 0.0
	numDuplicates = 0


	i = 0 
	while i < CSK_MODEL_VERTEX_COUNT:
		OBJMorphNormalizedNormals.append([])
		OBJMorphNormalizedNormals[i].append(0.0)
		OBJMorphNormalizedNormals[i].append(0.0)
		OBJMorphNormalizedNormals[i].append(0.0)
		i += 1
	i = 0 
	while i < CSK_MODEL_VERTEX_COUNT:
		###If the normal has not yet been set
		if (OBJMorphNormalizedNormals[i][0] == 0.0) and (OBJMorphNormalizedNormals[i][1] == 0.0) and (OBJMorphNormalizedNormals[i][2] == 0.0):
			numDuplicates = len(OBJVertexDuplicates[i])
			length = 0.0
			###Add current vertex data
			Vx = OBJMorphNormals[i][0]
			Vy = OBJMorphNormals[i][1]
			Vz = OBJMorphNormals[i][2]
			###Add any duplicate vertex data
			j = 0
			while j < numDuplicates:
				Vx += OBJMorphNormals[OBJVertexDuplicates[i][j]][0]
				Vy += OBJMorphNormals[OBJVertexDuplicates[i][j]][1]
				Vz += OBJMorphNormals[OBJVertexDuplicates[i][j]][2]
				j += 1
			###(+1) because the current vertex counts as 1, plus the number of duplicates
			Vx = Vx / (numDuplicates+1)
			Vy = Vy / (numDuplicates+1)
			Vz = Vz / (numDuplicates+1)
			length = math.sqrt( (Vx*Vx) + (Vy*Vy) + (Vz*Vz) )
			Vx = Vx / length
			Vy = Vy / length
			Vz = Vz / length

			###Set current vertex's data
			OBJMorphNormalizedNormals[i][0] = Vx
			OBJMorphNormalizedNormals[i][1] = Vy
			OBJMorphNormalizedNormals[i][2] = Vz
			###Set any duplicate vertex's data
			j = 0
			while j < numDuplicates:
				OBJMorphNormalizedNormals[OBJVertexDuplicates[i][j]][0] = Vx
				OBJMorphNormalizedNormals[OBJVertexDuplicates[i][j]][1] = Vy
				OBJMorphNormalizedNormals[OBJVertexDuplicates[i][j]][2] = Vz
				j += 1
		i += 1



###########################################
#   Calculate Tangent and Bitangent       #
###########################################
	tv = []
	bv = []
	V1 = 0
	V2 = 0
	V3 = 0
	x1 = 0
	x2 = 0
	y1 = 0
	y2 = 0
	z1 = 0
	z2 = 0
	s1 = 0
	s2 = 0
	t1 = 0
	t2 = 0
	r = 0
	dotNT = 0
	dotNB = 0
	TX = 0
	TY = 0
	TZ = 0
	BX = 0
	BY = 0
	BZ = 0
	TLength = 0
	BLength = 0


	i = 0
	while i < CSK_MODEL_VERTEX_COUNT:
		tv = [ 0.0 , 0.0 , 0.0 ]
		bv = [ 0.0 , 0.0 , 0.0 ]
		k = len(CSKVertexFaceIndex[i])
		j = 0
		###For each face that contains this vertex...
		while j < k:
			###V1 is the vertex INDEX of this vertex
			###V2 and V3 are the indexes of the other vertices in this face
			V1 = i
			if CSKFaces[CSKVertexFaceIndex[i][j]][0] == i:
				V2 = CSKFaces[CSKVertexFaceIndex[i][j]][1]
				V3 = CSKFaces[CSKVertexFaceIndex[i][j]][2]
			elif CSKFaces[CSKVertexFaceIndex[i][j]][1] == i:
				V2 = CSKFaces[CSKVertexFaceIndex[i][j]][0]
				V3 = CSKFaces[CSKVertexFaceIndex[i][j]][2]
			else:
				V2 = CSKFaces[CSKVertexFaceIndex[i][j]][0]
				V3 = CSKFaces[CSKVertexFaceIndex[i][j]][1]
			x1 = OBJMorphVertex[V2][0] - OBJMorphVertex[V1][0]
			x2 = OBJMorphVertex[V3][0] - OBJMorphVertex[V1][0]
			y1 = OBJMorphVertex[V2][1] - OBJMorphVertex[V1][1]
			y2 = OBJMorphVertex[V3][1] - OBJMorphVertex[V1][1]
			z1 = OBJMorphVertex[V2][2] - OBJMorphVertex[V1][2]
			z2 = OBJMorphVertex[V3][2] - OBJMorphVertex[V1][2]
			if CSKUV[i][0] > 0.5:
				s1 = ((1-(OBJMorphUV[V2][0]))*2) - ((1-(OBJMorphUV[V1][0]))*2)
				s2 = ((1-(OBJMorphUV[V3][0]))*2) - ((1-(OBJMorphUV[V1][0]))*2)
				t1 = (-OBJMorphUV[V2][1]) + OBJMorphUV[V1][1]
				t2 = (-OBJMorphUV[V3][1]) + OBJMorphUV[V1][1]
			else:
				s1 = (OBJMorphUV[V2][0]*2) - (OBJMorphUV[V1][0]*2)
				s2 = (OBJMorphUV[V3][0]*2) - (OBJMorphUV[V1][0]*2)
				t1 = (-OBJMorphUV[V2][1]) + OBJMorphUV[V1][1]
				t2 = (-OBJMorphUV[V3][1]) + OBJMorphUV[V1][1]
			r = 1.0 / ((s1*t2-s2*t1) + 0.00001)
			tv[0] += (t2*x1-t1*x2)*r
			tv[1] += (t2*y1-t1*y2)*r
			tv[2] += (t2*z1-t1*z2)*r
			bv[0] += (s1*x2-s2*x1)*r
			bv[1] += (s1*y2-s2*y1)*r
			bv[2] += (s1*z2-s2*z1)*r
			j += 1
		dotNT = (OBJMorphNormalizedNormals[i][0]*tv[0]) + (OBJMorphNormalizedNormals[i][1]*tv[1]) + (OBJMorphNormalizedNormals[i][2]*tv[2])
		TX = tv[0] - (OBJMorphNormalizedNormals[i][0] * dotNT)
		TY = tv[1] - (OBJMorphNormalizedNormals[i][1] * dotNT)
		TZ = tv[2] - (OBJMorphNormalizedNormals[i][2] * dotNT)

		###Misnomer here.. not length yet. The calculation really does call for the length, but it will
		###be squared anyway, no need to sqrt the length here yet
		TLength = (TX*TX) + (TY*TY) + (TZ*TZ)

		dotNB = (OBJMorphNormalizedNormals[i][0]*bv[0]) + (OBJMorphNormalizedNormals[i][1]*bv[1]) + (OBJMorphNormalizedNormals[i][2]*bv[2])
		dotTpB = (TX*bv[0]) + (TY*bv[1]) + (TZ*bv[2])
		###Handedness alway the opposite of this calculation
		BX = -(bv[0] - (OBJMorphNormalizedNormals[i][0] * dotNB) - (( TX * dotTpB) / TLength))
		BY = -(bv[1] - (OBJMorphNormalizedNormals[i][1] * dotNB) - (( TY * dotTpB) / TLength))
		BZ = -(bv[2] - (OBJMorphNormalizedNormals[i][2] * dotNB) - (( TZ * dotTpB) / TLength))


		BLength = math.sqrt( (BX*BX) + (BY*BY) + (BZ*BZ) )
		###Now make it the real length value
		TLength = math.sqrt(TLength)

		if BLength == 0:
			BX = 0.0
			BY = 0.0
			BZ = 0.0
		else:
			BX = BX / BLength
			BY = BY / BLength
			BZ = BZ / BLength

		if TLength == 0:
			TX = 0.0
			TY = 0.0
			TZ = 0.0
		else:		
			TX = TX / TLength
			TY = TY / TLength
			TZ = TZ / TLength


###TESTING - ELSE branch original commands
#		if CSKUV[i][0] > 0.5:
#			OBJMorphTangents.append( ( -TX , -TY , -TZ ) )
#			OBJMorphBitangents.append( ( BX , BY , BZ ) )
#		else:
		OBJMorphTangents.append( ( TX , TY , TZ ) )
		OBJMorphBitangents.append( ( BX , BY , BZ ) )
		i += 1

###########################################
#          Write morph data               #
###########################################
#Only the vertices that move in the morph are written into the CSK.
#Additionally, only displacement values are written for all of the data,
#not the 'straight' data from the morphed-model
#A bit-wise array created in the CSK indicates which vertices have moved
#0 = not moved (and not included in this morph's data)
#1 = moved (and data includded in this morph)
#There is one bit per vertex in the base model and all vertices are written
#or indicated in the same order as they apear in the base model data

###########################################
#          Write morph header             #
###########################################
	###Name of the morph
	scratchvalue += files[u+1][2]
	i = len(files[u+1][2])
	###Padding, avoiding use of Python string functions
	while i < 32:
		scratchvalue += "\x00"
		i += 1
	scratchvalue += struct.pack('<i', 72)
	#There should be the same number of vertices in each morph model as there are in the base model
	#This indicates the length of the bitwise array, not number of displaced verticies
	scratchvalue += struct.pack('<i', CSK_MODEL_VERTEX_COUNT)

	owfileCSK.write(scratchvalue)
	scratchvalue=''


###########################################
#    generate morph bitwise array         #
###########################################
#There was a mistake in the programming for this section. The
#array indicating which verticies displaced is bit-wise, but
#the CSK file structure requires one BYTE for each vertex.
#This is the reason for determining the number of bytes used
#for the 'bit-wise' array, then adding 0h00 up to the number
#of verticies in the full model.
#The 1's for displaced verticies are shifted into position in
#each byte using an inner loop that counts to 8.
	binnum = 0b00000000
	DispX = 0
	DispY = 0
	DispZ = 0


	i = 0
	###J will be the 8-bit shift counter. 
	###REMEMBER that the bytes' bits are in reverse order from how a peson would normally read. Vertex associations is
	###[ 7 6 5 4 3 2 1 0 ] , [ 15 14 13 12 11 10 9 8 ] , ...
	j = 0
	while i < CSK_MODEL_VERTEX_COUNT:
		###Determine if the vertex moved enough to be considered 'displaced'
		if (abs(OBJMorphVertex[i][0]-OBJVertex[i][0]) > 0.0001) or (abs(OBJMorphVertex[i][1]-OBJVertex[i][1]) > 0.0001) or (abs(OBJMorphVertex[i][2]-OBJVertex[i][2]) > 0.0001):
			###Vertex Position
			DispX = OBJMorphVertex[i][0]-OBJVertex[i][0]
			DispY = OBJMorphVertex[i][1]-OBJVertex[i][1]
			DispZ = OBJMorphVertex[i][2]-OBJVertex[i][2]
			###Avoid creating very small displacement values. X may have moved enough
			###to 'displace' the vertex, but Y and Z moved very little. Rather than have
			###Y or Z's movement be 0.1E-12 or something, just make it zero.
			if abs(DispX) < 0.0001:
				DispX = 0
			if abs(DispY) < 0.0001:
				DispY = 0
			if abs(DispZ) < 0.0001:
				DispZ = 0
			###TESTING - see why there are some 0,0,0 morph vertices
			if DispX == 0 and DispY == 0 and DispZ ==0:
				print OBJMorphVertex[i][0]-OBJVertex[i][0]
				print OBJMorphVertex[i][1]-OBJVertex[i][1]
				print OBJMorphVertex[i][2]-OBJVertex[i][2]
			OBJMorphVertexDisplacement.append( ( DispX,DispY,DispZ ) )

			###Bitangent
			DispX = OBJMorphBitangents[i][0]-OBJBitangents[i][0]
			DispY = OBJMorphBitangents[i][1]-OBJBitangents[i][1]
			DispZ = OBJMorphBitangents[i][2]-OBJBitangents[i][2]
			if abs(DispX) < 0.0001:
				DispX = 0
			if abs(DispY) < 0.0001:
				DispY = 0
			if abs(DispZ) < 0.0001:
				DispZ = 0
			OBJMorphBitangentsDisplacement.append( ( DispX,DispY,DispZ ) )

			###Tangent
			DispX = OBJMorphTangents[i][0]-OBJTangents[i][0]
			DispY = OBJMorphTangents[i][1]-OBJTangents[i][1]
			DispZ = OBJMorphTangents[i][2]-OBJTangents[i][2]
			if abs(DispX) < 0.0001:
				DispX = 0
			if abs(DispY) < 0.0001:
				DispY = 0
			if abs(DispZ) < 0.0001:
				DispZ = 0
			OBJMorphTangentsDisplacement.append( ( DispX,DispY,DispZ ) )

			###Normal
			DispX = OBJMorphNormalizedNormals[i][0]-OBJNormalizedNormals[i][0]
			DispY = OBJMorphNormalizedNormals[i][1]-OBJNormalizedNormals[i][1]
			DispZ = OBJMorphNormalizedNormals[i][2]-OBJNormalizedNormals[i][2]
			if abs(DispX) < 0.0001:
				DispX = 0
			if abs(DispY) < 0.0001:
				DispY = 0
			if abs(DispZ) < 0.0001:
				DispZ = 0
			OBJMorphNormalizedNormalsDisplacement.append( ( DispX,DispY,DispZ ) )
			
			###Since this vertex displaced, make its bit "1"
			binnum = binnum | 0b10000000
			j += 1
		else:
			###If the vertex did not displace, it's bit position in the byte will be left as
			###zero, but it still "has" a position and the byte is still shifted
			j += 1
		###If 8 bits have been pressed into the binnum variable, write it out, prepare the next. Remember, 
		###counting from 0, not 1.
		if j > 7:
			scratchvalue += struct.pack('<B', binnum)
			j = 0
			binnum = 0b00000000
		binnum = binnum >> 1
		i += 1


	###Now, need to fill in the rest of the CSK_MODEL_VERTEX_COUNT bytes of space with padding.
	###Last few bits of the last useful byte need to be filled in, also. In the loop above, one
	###last pass at j=7 should be done to get the last bit set, here don't want that.
	while j < 7:
		binnum = binnum >> 1
		j += 1
	scratchvalue += struct.pack('<B', binnum)
	i = len(scratchvalue)
	while i < CSK_MODEL_VERTEX_COUNT:
		scratchvalue += "\x00"
		i += 1

###########################################
#    generate morph vertex array          #
###########################################
	###size of the morph vertex array is the number of entries in it (verties moved)
	###times the length of each vertex entry (72 bytes)
	scratchvalue += struct.pack('<i', len(OBJMorphVertexDisplacement)*72)

	j = len(OBJMorphVertexDisplacement)
	i = 0
	while i < j:
		scratchvalue += struct.pack('<f', OBJMorphVertexDisplacement[i][0]) + struct.pack('<f', OBJMorphVertexDisplacement[i][1]) + struct.pack('<f', OBJMorphVertexDisplacement[i][2]) + struct.pack('<f', OBJMorphNormalizedNormalsDisplacement[i][0]) + struct.pack('<f', OBJMorphNormalizedNormalsDisplacement[i][1]) + struct.pack('<f', OBJMorphNormalizedNormalsDisplacement[i][2]) + struct.pack('<f', OBJMorphTangentsDisplacement[i][0]) + struct.pack('<f', OBJMorphTangentsDisplacement[i][1]) + struct.pack('<f', OBJMorphTangentsDisplacement[i][2]) + struct.pack('<f', OBJMorphBitangentsDisplacement[i][0]) + struct.pack('<f', OBJMorphBitangentsDisplacement[i][1]) + struct.pack('<f', OBJMorphBitangentsDisplacement[i][2]) + struct.pack('<f', OBJMorphTangentsDisplacement[i][0]) + struct.pack('<f', OBJMorphTangentsDisplacement[i][1]) + struct.pack('<f', OBJMorphTangentsDisplacement[i][2]) + struct.pack('<f', OBJMorphBitangentsDisplacement[i][0]) + struct.pack('<f', OBJMorphBitangentsDisplacement[i][1]) + struct.pack('<f', OBJMorphBitangentsDisplacement[i][2])

		i += 1

###########################################
#           Write out morph               #
###########################################
#Finally, oi. Time for a break.. tried adding cinnamon and clove powder
#to the ground coffee, pretty good.
	owfileCSK.write(scratchvalue)
	scratchvalue = ''

	###Close morph OBJ file since we're done
	orfile.close()
	u += 1
################################################
################################################

print ": done"

################################################
################################################


		










###########################################
#      Write out file section size        #
###########################################
#Go back and write in the size for the radeon
#section now that it's complete

tempLoc = owfileCSK.tell()
owfileCSK.seek(NewCSKwriteFileSectionSizeLoc)
###Where we were in the current file minus where the section began (section began
###directly AFTER the size integer)
scratchvalue += struct.pack('<i', tempLoc - (NewCSKwriteFileSectionSizeLoc + 4) )
owfileCSK.write(scratchvalue)
scratchvalue = ''
owfileCSK.seek(tempLoc)







###########################################
#    Generate Edge defenition sections    #
###########################################
#The actual CSK file's edge defenitions are in half edges
# A half edge means that both sides of a line connecting two vertices are labeled differently
#           14 
#    * ----------- *    - Top part of line between vertices is edge 14, bottom is edge 456
#           456           both sides belong to different mesh faces
#For each edge of a face as indicated in the VertexEdgeArray, the CSK indicates which
#half-edge is it's twin (on the other side) and which edge is it's neighbor (just the
#'next' edge in that face, really, so +1 to that edges edge number).
#Every edge not on a UV seam will have a twin (will create two half-edges). Realize
#that 14 and 456 above are in two different faces, and each face will have indicated
#this edge, so the edge will appear in VertexEdgeArray twice (with the same vertex
#indexes specifying the edge).
#
#There are three edge sections. The first seems to define half-edge association,
#The second was a face-to-edge defenition for the geforce section's model (so
#ignored and made blank here), the third section is the face-to-edge for the
#radeon section's model.
#
#
#Again, finally, since the mesh cannot be modified, just copy the arrays


print "Generating CSK edge definitions . . .",

###########################################
#           EdgeConnectivity              #
###########################################
entryCount = 0

###Write section header
scratchvalue += "EdgeConnectivity"
scratchvalue += "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
###Write section size. number of edges = num of faces * 3. Each entry is 4 ints = 16 bytes
###+4 for count number which is part of the section size
scratchvalue += struct.pack( '<i', (((CSK_FACE_COUNT*3)*16)+4) )

entryCount = CSK_FACE_COUNT*3

i = 0
while i < entryCount:
	scratchvalue += struct.pack('<4i', CSKEdgeConn[i][0], CSKEdgeConn[i][1], CSKEdgeConn[i][2], CSKEdgeConn[i][3] )
	i += 1


###########################################
#           Connectivity96Const           #
###########################################
#This section is skipped. It must exist, but is made 0 size

scratchvalue += "Connectivity96Const"
scratchvalue += "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
###The section size is 4 bytes, since the size of the sub-array count
###directly following is part of the section size
scratchvalue += struct.pack('<i', 4)
scratchvalue += struct.pack('<i', 0)



###########################################
#           Connectivity256Const           #
###########################################
scratchvalue += "Connectivity256Const"
scratchvalue += "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
###The size of this section. It will be the size of each array entry (24 bytes)
###times the number of entries (the number of model faces) plus the next count number
###(another 4 bytes)
scratchvalue += struct.pack('<i', ((24*CSK_FACE_COUNT)+4) )
###number of entries in this array (The number of edges)
scratchvalue += struct.pack('<i', CSK_FACE_COUNT)

i = 0
while i < CSK_FACE_COUNT:
	scratchvalue += struct.pack('<8i', CSKConn256[i][0], CSKConn256[i][1], CSKConn256[i][2], CSKConn256[i][3], CSKConn256[i][4], CSKConn256[i][5], CSKConn256[i][6], CSKConn256[i][7] )
	i += 1

owfileCSK.write(scratchvalue)
scratchvalue = ''


################################################
################################################

print " done"

################################################
################################################


owfileCSK.close()










